# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Nark User Configurable Settings"""

import datetime
import os
from collections import namedtuple

from gettext import gettext as _

from config_decorator import section, ConfigDecorator

from ..helpers.app_dirs import NarkAppDirs

from .log_levels import get_log_level_safe, must_verify_log_level

__all__ = (
    'REGISTERED_BACKENDS',
    'ConfigRoot',
    'decorate_config',
    # PRIVATE:
    # 'NarkConfigurableDb',
    # 'NarkConfigurableDev',
    # 'NarkConfigurableTime',
)

BackendRegistryEntry = namedtuple(
    'BackendRegistryEntry', ('verbose_name', 'store_class'),
)

REGISTERED_BACKENDS = {
    'sqlalchemy': BackendRegistryEntry(
        'SQLAlchemy',
        'nark.backends.sqlalchemy.SQLAlchemyStore',
    ),
}


# ***
# *** Top-level, root config object.
# ***

@section(None)
class ConfigRoot(object):
    pass


# ***
# *** Backend (nark) Config.
# ***


@ConfigRoot.section('db')
class NarkConfigurableDb(object):
    """"""

    def __init__(self, *args, **kwargs):
        # Cannot super because @section decorator makes NarkConfigurableDb
        # an object instance of a different class type, so there is really
        # no NarkConfigurableDb class. It's been replaced by/monkey patched
        # into an object instance of the decorator class. As such, calling
        # super here, without the correct class, would raise.
        # DEATH: super(NarkConfigurableDb, self).__init__(*args, **kwargs)
        pass

    # ***

    @property
    @ConfigRoot.setting(
        # Just showing off: How to deliberately specify the setting name,
        # otherwise it defaults to the function name.
        # (lb): Was called 'store' in hamster-lib, but 'orm' descriptiver.
        name='orm',  # I.e., instead if `def orm(self)`.
        choices=REGISTERED_BACKENDS,
    )
    def store_default(self):
        # HINT: The doc string here can be used in lieu of specifying in @setting.
        """ORM used by dob to interface with the DBMS. Most likely ‘sqlalchemy’."""
        # HINT: The property return value is the default for the setting.
        # HINT: The type of this value determines the setting type, too.
        return 'sqlalchemy'

    # ***

    @property
    @ConfigRoot.setting(
        _("Database management system used to manage your data."
            " Most likely ‘sqlite’."),
    )
    def engine(self):
        return 'sqlite'

    # ***

    @property
    @ConfigRoot.setting(
        # (lb): It's possible db.path could apply to something other than
        # SQLite that also saves to a single file. But until I hear of someone
        # who actually uses something other than SQLite, let's make the
        # documentation more precise, and specifically say applies to SQLite.
        # Also because I haven't tested anything other than SQLite.
        _("Path to SQLite database file"
            " (for ‘sqlite’ db.engine)."),
        # The db.path only applies for 'sqlite' DBMS.
        # But we won't set ephemeral or hidden, because user should still see in
        # config, so they more easily understand how to change DBMS settings.
    )
    def path(self):
        if NarkAppDirs.APP_DIRS is None:
            # Happens when code is sourced, before NarkAppDirs() created.
            return ''
        return os.path.join(
            NarkAppDirs.APP_DIRS.user_data_dir,
            # MAYBE: Rename? 'nark.sqlite'?? or 'hamster.sqlite'??
            # FIXME: Make this a package const rather than inline literal.
            #        (Maybe on Config refactor how to do so will be evident.)
            'dob.sqlite',
        )

    # ***

    # The 5 settings -- db.host, db.port, db.name, db.user, and db.password --
    # apply when db.engine != 'sqlite'. Otherwise, if sqlite, only db.path used.

    @property
    @ConfigRoot.setting(
        _("Host name of the database server"
            " (for non-‘sqlite’ db.engine)."),
    )
    def host(self):
        return ''

    @property
    @ConfigRoot.setting(
        _("Port number on which the server is listening"
            " (for non-‘sqlite’ db.engine)."),
    )
    def port(self):
        return ''

    @property
    @ConfigRoot.setting(
        _("The database name (for non-‘sqlite’ db.engine)."),
    )
    def name(self):
        return ''

    @property
    @ConfigRoot.setting(
        _("The database user (for non-‘sqlite’ db.engine)."),
    )
    def user(self):
        return ''

    @property
    @ConfigRoot.setting(
        _("The database password (non-‘sqlite’)."
            " WARNING: This setting is potentially unsafe!"),
    )
    def password(self):
        return ''


# ***

@ConfigRoot.section('dev')
class NarkConfigurableDev(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        _("The log level for library (nark) squaller"
            " (using Python logging library levels)"),
        validate=must_verify_log_level,
        conform=get_log_level_safe,
    )
    def lib_log_level(self):
        return 'WARNING'

    @property
    @ConfigRoot.setting(
        _("The log level for database (SQL) squaller"
            " (using Python logging library levels)"),
        validate=must_verify_log_level,
        conform=get_log_level_safe,
    )
    def sql_log_level(self):
        return 'WARNING'


# ***

def _strptime_day_start(day_start_text):

    def _parse_day_start():
        if isinstance(day_start_text, datetime.time):
            return day_start_text
        if not day_start_text:
            return datetime.time(0, 0, 0)
        # Returns a datetime.time or raises a ValueError.
        return _must_parse_text()

    def _must_parse_text():
        try:
            return datetime.datetime.strptime(day_start_text, '%H:%M:%S').time()
        except ValueError:
            warn_invalid()

    def warn_invalid():
        msg = _(" (Expected '%H:%M:%S'-formatted time of day, not ‘{}’.)".format(
            day_start_text,
        ))
        raise ValueError(msg)

    return _parse_day_start()


def _validate_day_start(day_start_text):
    """Returns True if the day_start_text is valid, or raises ValueError."""
    _strptime_day_start(day_start_text)
    return True


# ***

@ConfigRoot.section('time')
class NarkConfigurableTime(object):
    """"""

    def __init__(self, *args, **kwargs):
        pass

    # ***

    @property
    @ConfigRoot.setting(
        # allow_momentaneous indictes if 0-length duration Facts allowed,
        #   e.g., start == end.
        # Here's a little vocabulary lesson:
        #      Fugacious: "lasting a short time"
        #     Evanescent: "tending to vanish like vapor"
        #   Momentaneous: "characterizing action begun, terminated in an instant"
        #   See also:     Vacant, Fleeting, Transient.
        _("If True, lets you save Facts that start and end"
            " at the exact same time, i.e., zero-length Facts."),
        # FIXME/2019-11-18: (lb): Hidden for now. Still testing.
        # ! - Might be better to default True.
        hidden=True,
    )
    def allow_momentaneous(self):
        return False

    # ***

    @property
    @ConfigRoot.setting(
        # hamster-lib described day_start as simply "can be used to specify
        # default start time". I.e., when you create a new Fact, if there
        # was no earlier completed Fact from that same day whose end time
        # could be used as the default start time, hamster would set the
        # start time to day_start. This could be useful if you did not dob
        # all 24 hours of your day, and your day tended to start at the same
        # time. E.g., if you always got to work at 7am, you could set day_start
        # to 7am, and when you'd create that first Fact in the morning, the
        # start time would default to 7am. Pretty straightforward. In dob it
        # works the same, except dob defaults the start time to the end time
        # of the most recent completed Fact, even if it's days or years before,
        # because dob discourages users from leaving gaps in their facts.
        #   The day_start serves additional purposes in dob, too. If the user
        # specifies a datetime option using just a date, but no time, dob either
        # defaults to midnight of that day, or it uses day_start. Also, when
        # generating daily reports, dob either groups Facts each day from
        # midnight to midnight, or from between day_start on consecutive days.
        # (lb): This help is tricky to get right. This iteration feels okay:
        _("Default start time for grouping by days, and for dates with no time."),
        conform=_strptime_day_start,
        # Because validate called after conform, and because _validate_day_start
        # calls same parser method as _strptime_day_start, setting validate here
        # would be redundant, so skipping:
        #     validate=_validate_day_start,
    )
    def day_start(self):
        # Same default as in Legacy Hamster, midnight, a sextuple zed double colon.
        return '00:00:00'

    # ***

    @property
    @ConfigRoot.setting(
        _("A Fact being saved must have a duration of at least this many seconds."),
    )
    def fact_min_delta(self):
        # (lb): Disable this by default; I've never liked this logic!
        #   In Legacy Hamster: 60, i.e., facts must be 1 minute apart!
        #   In Modern Hamster (nark), you can make facts every seconds,
        #     or every millisecond, we don't care, so long as they do
        #     not overlap!
        return 0

    # ***

    # FIXME/2020-01-28: (lb): This never got implemented...

    @property
    @ConfigRoot.setting(
        _("Not Implemented."),
        ephemeral=True,
        hidden=True,
    )
    def tz_aware(self):
        return False

    @property
    @ConfigRoot.setting(
        _("Not Implemented."),
        ephemeral=True,
        hidden=True,
    )
    def default_tzinfo(self):
        return ''


# ***

def decorate_config(config):
    """Wraps or ensures the supplied config dict is a nark config decorator.

    Note: This is mostly for the ``nark/tests``, as the dob client will
    explicitly prepare a decorated config and pass that to nark. But to
    make it easier to write fixtures for ``nark/tests``, this method
    accepts a dictionary and returns a decorated config.
    """
    if isinstance(config, ConfigDecorator):
        return config
    # Caution: Remember that ConfigRoot is a weird Singleton.
    # FIXME/2020-01-07: (lb): Can probably put ConfigRoot et al in a getter
    #                         so that it doesn't have to be a module global.
    config_root = ConfigRoot
    config_root.forget_config_values()
    config_root.update(config)
    return config_root

# ***




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

from gettext import gettext as _

import logging
import os
import re

import lazy_import

from .parse_errors import (
    ParserException,
    ParserInvalidDatetimeException,
    ParserMissingActivityException,
    ParserMissingDatetimeOneException,
    ParserMissingDatetimeTwoException,
    ParserMissingSeparatorActivity
)
from .parse_time import HamsterTimeSpec, parse_datetime_iso8601

# Profiling: `import dateparser` takes ~ 0.2 seconds.
dateparser = lazy_import.lazy_module('dateparser')


__all__ = (
    'parse_factoid',
    'Parser',
)


# FIXME/MAYBE: (lb): New pattern? Can modules just get the logger here?
#   Or should we make a top-level module that just returns this? Probably
#   the latter, so we're not hard-coding 'nark.log' everywhere.
logger = logging.getLogger('nark.log')


# FIXME: (lb): What's the best way to handle module-scope vars like this?
#        Should this be from config?
#        From a "globals" module?
#        From a function-scoped sub-function?
#        Or is here fine?
DATE_TO_DATE_SEPARATORS__RAW = ['to', 'until', '-']


FACT_METADATA_SEPARATORS = [",", ":"]


# Map time_hint to minimum and maximum datetimes to seek.
TIME_HINT_CLUE = {
    'verify_start': (1, 2),  # end time is optional.
    'verify_end': (1, 1),  # exactly one is required.
    'verify_then': (0, 2),  # both times are optional.
    'verify_still': (0, 2),  # both times are optional.
    'verify_none': (0, 0),  # none is none is all alone.
    'verify_both': (2, 2),  # and both is both times two.
}


class Parser(object):
    """FIXME"""

    ACTEGORY_SEP = '@'

    RE_DATE_TO_DATE_SEP = None
    RE_SPLIT_CAT_AND_TAGS = None
    RE_SPLIT_TAGS_AND_TAGS = None

    def __init__(self):
        self.reset()

    def reset(self):
        self.reset_rules()
        self.reset_result()

    def reset_rules(self):
        self.raw = None
        self.flat = None
        self.rest = None

        self.time_hint = None
        self.re_item_sep = None
        self.hash_stamps = None
        self.lenient = None
        self.local_tz = None
        self.skip_dateparser = False

    def reset_result(self):
        self.datetime1 = None
        self.datetime2 = None
        self.raw_datetime1 = None
        self.raw_datetime2 = None
        self.type_datetime1 = None
        self.type_datetime2 = None
        self.activity_name = None
        self.category_name = None
        self.tags = None
        self.description = None
        self.warnings = []

    def __str__(self):
        return (
            'raw: {}'
            ' / flat: {}'
            ' / rest: {}'

            ' / time_hint: {}'
            ' / re_item_sep: {}'
            ' / hash_stamps: {}'
            ' / lenient: {}'
            ' / local_tz: {}'
            ' / skip_dateparser: {}'

            ' / datetime1: {}'
            ' / datetime2: {}'
            ' / raw_datetime1: {}'
            ' / raw_datetime2: {}'
            ' / type_datetime1: {}'
            ' / type_datetime2: {}'
            ' / activity_name: {}'
            ' / category_name: {}'
            ' / tags: {}'
            ' / description: {}'
            ' / warnings: {}'
            .format(
                self.raw,
                self.flat,
                self.rest,

                self.time_hint,
                self.re_item_sep,
                self.hash_stamps,
                self.lenient,
                self.local_tz,
                self.skip_dateparser,

                self.datetime1,
                self.datetime2,
                self.raw_datetime1,
                self.raw_datetime2,
                self.type_datetime1,
                self.type_datetime2,
                self.activity_name,
                self.category_name,
                self.tags,
                self.description,
                self.warnings,
            )
        )

    # **************************************
    # *** All the patterns we're gonna need!
    # **************************************

    def setup_patterns(self):
        self.re_setup_datatimes_separator()
        self.re_setup_category_and_tags()
        self.re_setup_tags_upon_tags()

    def re_setup_datatimes_separator(self):
        Parser.RE_DATE_TO_DATE_SEP = re.compile(
            r'\s(to|until|\-)\s|(?<=\d)(\-)(?=\d)'
        )

    def re_setup_category_and_tags(self):
        # FIXME/2018-05-15: (lb): Should #|@ be settable, like the other
        # two (DATE_TO_DATE_SEPARATORS and FACT_METADATA_SEPARATORS)?
        # Or does that make maintaining the parser that much harder?
        # HINT: Matches space(s) followed by hash.
        #   On split, removes whitespace (because matched).
        #   - First split element may be empty string.
        #   - Final split element may have trailing spaces.
        Parser.RE_SPLIT_CAT_AND_TAGS = re.compile(
            r'\s+[{hash_stamps}](?=\S)'
            .format(hash_stamps=self.hash_stamps)
        )

    def re_setup_tags_upon_tags(self):
        # HINT: Matches only on a hash starting the string.
        #   On split, leaves trailing spaces on each element.
        #   - First split element may be whitespace string.
        Parser.RE_SPLIT_TAGS_AND_TAGS = re.compile(
            r'(?<!\S)[{hash_stamps}](?=\S)'
            .format(hash_stamps=self.hash_stamps)
        )

    # **************************************
    # *** dissect_raw_fact: Main class entry
    # **************************************

    def dissect_raw_fact(self, *args, **kwargs):
        """
        FIXME: Document
        """
        # The raw_fact is a tuple. The user can use quotes or not, and it's
        # up to us to figure things out.

        self.prepare_parser(*args, **kwargs)

        err = None

        try:
            self.parse()
        except ParserException as perr:
            err = perr
            if not self.lenient:
                raise

        return err

    # ************************************
    # *** Helper fcns for dissect_raw_fact
    # ************************************

    def parse(self):
        self.reset_result()
        try:
            # If the date(s) are ISO 8601, find 'em fast.
            after_datetimes = self.parse_datetimes_easy()
            # Datetime(s) were 8601 (code did not raise),
            # so now look for the '@' and set the activity.
            rest_after_act, expect_category = self.lstrip_activity(after_datetimes)
        except ParserException:
            self.reset_result()
            rest_after_act, expect_category = self.parse_datetimes_hard()
        if expect_category:
            self.parse_cat_and_remainder(rest_after_act)
        else:
            self.parse_tags_and_remainder(rest_after_act)
        # Validation.
        self.hydrate_datetimes()
        if not self.activity_name:
            self.raise_missing_activity()
        # We could raise on missing-category; or
        #   maybe caller can deduce, so don't.
        # Don't care if tags or description are empty.

    def prepare_parser(self, *args, **kwargs):
        self.setup_rules(*args, **kwargs)
        self.setup_patterns()

    def setup_rules(
        self,
        factoid,
        time_hint='verify_none',
        separators=None,
        hash_stamps=None,
        lenient=False,
        # FIXME/2018-05-22 20:42: (lb): Implement: tz_local
        local_tz=None,  # Default to None, i.e., naive
    ):
        if isinstance(factoid, str):
            # Path from tests/nark/test_objects.py, but not from dob.
            factoid = (factoid,)
        else:
            # The user can get here on an empty --ask, e.g.,
            #   ``nark on --ask``
            factoid = factoid or ('',)

        # Parse a flat copy of the args.
        full = ' '.join(factoid)
        parts = full.split(os.linesep, 1)
        flat = parts[0]
        more_description = '' if len(parts) == 1 else parts[1].strip()

        # Items are separated by any one of the separator(s)
        # not preceded by whitespace, and followed by either
        # whitespace, or end of string/before newline.
        if not separators:
            separators = FACT_METADATA_SEPARATORS
        assert len(separators) > 0
        sep_group = '|'.join(separators)
        # Gobble whitespace as part of separator, to make it easier to pull
        # data apart and then part it back together if we need. E.g., if user
        # puts description on same line as meta data, and if description contains
        # separators, we'll split the line first to parse out the meta data, and
        # then we'll put it back together, so if a separator is part of the
        # description, we want to be sure to retain the whitespace around the
        # separator if we have to patch the description back together from its
        # parts that did not turn out to be meta data (like #tags).
        # This is how parser originally split, leaving whitespace in the last part:
        #   # C✗P✗: re.compile('(?:,|:)(?=\\s|$)')
        #   ..._sep = re.compile(r'({})(?=\s|$)'.format(sep_group))
        # We can pull whitespace into the separator with two Levenshtein moves.
        re_item_sep = re.compile(r'({}(?=\s+|$))'.format(sep_group))

        if not hash_stamps:
            hash_stamps = '#@'

        self.reset()
        self.raw = factoid
        self.flat = flat
        self.rest = more_description
        self.time_hint = time_hint
        self.re_item_sep = re_item_sep
        self.hash_stamps = hash_stamps
        self.lenient = lenient
        self.local_tz = local_tz

    def parse_datetimes_easy(self):
        rest = self.flat
        if self.time_hint == 'verify_end':
            rest, _sep = self.must_parse_datetime_from_rest(rest, 'datetime2')
        elif self.time_hint != 'verify_none':
            minmax = TIME_HINT_CLUE[self.time_hint]
            rest = self.parse_datetimes_easy_both(rest, minmax)
        # else, time_hint == 'verify_none', so rest is rest.
        return rest

    def parse_datetimes_easy_both(self, rest, minmax):
        try:
            rest, sep = self.must_parse_datetime_from_rest(rest, 'datetime1')
        except ParserMissingDatetimeOneException:
            if minmax[0] == 0:
                return rest
            raise
        strictly_two = (minmax[0] == 2)
        # If sep is nonempty (e.g., ':', or ','), do not expect datetime2.
        if not sep:
            # The next token in rest could be the "to"/"until"/"-" sep.
            parts = Parser.RE_DATE_TO_DATE_SEP.split(rest, 1)
            # ... however, the RE_DATE_TO_DATE_SEP regex matches anywhere in line.
            # So verify that first part of split is empty, otherwise to/until sep
            # does not start the rest of the factoid.
            if (parts[0].strip() == '') and (len(parts) > 1):
                # There are four capture groups:
                #   1. The stuff before the times;
                #   2. The 'to'/'until'/'-', if separator matched on word boundary,
                #        e.g., "X to Y"; otherwise, None.
                #   3. The '-', if separator matched on digit boundary,
                #        e.g., "9-5"; otherwise, None.
                #   4. The stuff after the (2.) or (3.) separator.
                assert len(parts) == 4
                assert (parts[1] is None) ^ (parts[2] is None)
                separator = parts[1] or parts[2]
                assert separator in DATE_TO_DATE_SEPARATORS__RAW
                after_dt2, _sep = self.must_parse_datetime_from_rest(
                    parts[3], 'datetime2', ok_if_missing=(not strictly_two),
                )
                if after_dt2 is not None:
                    rest = after_dt2
                # else, was not a datetime, so re-include DATE_TO_DATE_SEPARATOR.
            elif strictly_two:
                self.raise_missing_datetime_two()
        elif strictly_two:
            self.raise_missing_datetime_two()
        return rest

    def parse_datetimes_hard(self):
        expect_category = True
        if self.time_hint == 'verify_none':
            # There is no datetime(s). Just the act@cat.
            rest_after_act, expect_category = self.lstrip_activity(self.flat)
        else:
            minmax = TIME_HINT_CLUE[self.time_hint]
            rest_after_act = self.lstrip_datetimes(minmax)
        return (rest_after_act, expect_category)

    def lstrip_datetimes(self, minmax):
        two_is_okay = (minmax[1] == 2)
        strictly_two = (minmax[0] == 2)
        parts = self.lstrip_datetimes_delimited()
        datetimes_and_act, datetimes, rest_after_act = parts
        if datetimes:
            self.must_parse_datetimes_known(datetimes, two_is_okay, strictly_two)
            # We've processed datetime1, datetime2, and activity_name.
        else:
            # The user did not delimit the datetimes and the activity.
            # See if the user specified anything magically, otherwise, bye.
            self.must_parse_datetimes_magic(datetimes_and_act, two_is_okay, strictly_two)
        return rest_after_act

    def lstrip_datetimes_delimited(self):
        # If user wants to use friendly datetimes, they need to delimit, e.g.:
        #   `nark yesterday until today at 3 PM, act @ cat # tag 1, descrip`
        # Note that the special token 'now' could be considered okay:
        #   `nark yesterday at 3 PM until now act @ cat # tag 1 "descrip"`
        # First look for the activity@category separator, '@'. This is a simple
        # find (index) because we insist that neither the datetime, nor the datetimes
        # sep, include the `@` symbol; and that @tags follow the activity@category.
        act_cat_sep_idx = self.must_index_actegory_sep(self.flat, must=True)
        # Next, split the raw factoid into two: datetime(s) and activity; and the rest.
        datetimes_and_act = self.flat[:act_cat_sep_idx]
        rest_after_act = self.flat[act_cat_sep_idx + 1:]
        # Determine if the user delimited the datetime(s) from the activity
        # using, e.g., a comma, ',' (that follows not-whitespace, and is
        # followed by whitespace/end-of-string). (Note that ':' can also be
        # used as the delimiter -- even though it's used to delimit time --
        # because the item separator must be the last character of a word.)
        parts = self.re_item_sep.split(datetimes_and_act, 1)
        if len(parts) == 3:
            datetimes = parts[0]
            # parts[1] is the sep, e.g., ',' or ':'.
            self.activity_name = parts[2]
        else:
            assert len(parts) == 1
            datetimes = None
        return (datetimes_and_act, datetimes, rest_after_act)

    def must_index_actegory_sep(self, part, must=True):
        try:
            # Find the first '@' in the raw, flat factoid.
            return part.index(Parser.ACTEGORY_SEP)
        except ValueError:
            if must:
                # It's only mandatory that we find an activity if the
                # datetimes are not ISO 8601 (otherwise we cannot tell
                self.raise_missing_separator_activity()
            return -1

    # *** 1: Parse datetime(s) and activity.

    def must_parse_datetimes_known(self, datetimes, two_is_okay, strictly_two):
        assert self.raw_datetime1 is None
        assert self.raw_datetime2 is None
        assert two_is_okay or (not strictly_two)

        if two_is_okay:
            # Look for separator, e.g., " to ", or " until ", or " - "/"-", etc.
            parts = Parser.RE_DATE_TO_DATE_SEP.split(datetimes, 1)
            if len(parts) > 1:
                assert len(parts) == 4  # middle 2 parts are the separator
                assert (parts[1] is None) ^ (parts[2] is None)
                separator = parts[1] or parts[2]
                assert separator in DATE_TO_DATE_SEPARATORS__RAW

                self.raw_datetime1 = parts[0]  # first datetime
                self.raw_datetime2 = parts[3]  # other datetime
            elif strictly_two:
                self.raise_missing_datetime_two()

        if not self.raw_datetime1:
            # Maybe not two_is_okay; definitely not strictly_two.
            if self.time_hint == 'verify_end':
                self.datetime1 = ''
                self.raw_datetime2 = datetimes
            else:
                self.raw_datetime1 = datetimes
                self.datetime2 = ''

    def must_parse_datetimes_magic(self, datetimes_and_act, two_is_okay, strictly_two):
        assert self.raw_datetime1 is None
        assert self.raw_datetime2 is None
        assert two_is_okay or (not strictly_two)

        if two_is_okay:
            # Look for separator, e.g., " to ", or " until ", or " - ", etc.
            parts = Parser.RE_DATE_TO_DATE_SEP.split(datetimes_and_act, 1)
            if len(parts) > 1:
                assert len(parts) == 4
                assert (parts[1] is None) ^ (parts[2] is None)
                separator = parts[1] or parts[2]
                assert separator in DATE_TO_DATE_SEPARATORS__RAW

                self.raw_datetime1 = parts[0]
                dt_and_act = parts[3]
                dt_attr = 'datetime2'
            elif strictly_two:
                self.raise_missing_datetime_two()

        if not self.raw_datetime1:
            dt_and_act = datetimes_and_act
            if self.time_hint == 'verify_end':
                dt_attr = 'datetime2'
            else:
                dt_attr = 'datetime1'

        rest, _sep = self.must_parse_datetime_from_rest(dt_and_act, dt_attr)
        self.activity_name = rest

    def must_parse_datetime_from_rest(
        self, datetime_rest, datetime_attr, ok_if_missing=False,
    ):
        assert datetime_attr in ['datetime1', 'datetime2']
        assert not ok_if_missing or datetime_attr == 'datetime2'
        # See if datetime: '+/-n' mins, 'nn:nn' clock, or ISO 8601.
        dt, type_dt, sep, rest = HamsterTimeSpec.discern(datetime_rest)
        if dt is not None:
            assert type_dt
            if type_dt == 'datetime':
                self.warn_if_datetime_missing_clock_time(dt, rest)
                dt = parse_datetime_iso8601(dt, must=True, local_tz=self.local_tz)
            # else, relative time, or clock time; let caller handle.
            setattr(self, datetime_attr, dt)
            # Set either 'type_datetime1' or 'type_datetime2'.
            # (NOTE: (lb): No caller actually uses type_dt.)
            setattr(self, 'type_{}'.format(datetime_attr), type_dt)
        elif datetime_attr == 'datetime1':
            self.raise_missing_datetime_one()
        elif not ok_if_missing:
            assert datetime_attr == 'datetime2'
            self.raise_missing_datetime_two()
        else:
            rest = None
        return rest, sep

    def warn_if_datetime_missing_clock_time(self, raw_dt, after_dt):
        if HamsterTimeSpec.has_time_of_day(raw_dt):
            return
        # NOTE: re.match checks for a match only at the beginning of the string.
        looks_like_clock_abbrev = re.match(r'\s*(\d:\d{2}|\d{3})(\s+|$)', after_dt)
        warn_msg = _('The identified datetime is missing the time of day.')
        if looks_like_clock_abbrev:
            warn_msg += _(
                ' Perhaps those three digits after the date'
                ' should be a 4-digit clocktime?'
            )
        else:
            warn_msg += _(' Is that what you wanted? (Probably not!)')
        self.warnings.append(warn_msg)

    def lstrip_activity(self, act_and_rest):
        act_cat_sep_idx = self.must_index_actegory_sep(act_and_rest, must=False)
        if act_cat_sep_idx >= 0:
            just_the_activity = act_and_rest[:act_cat_sep_idx]
            rest_after_act = act_and_rest[act_cat_sep_idx + 1:]
            expect_category = True
            self.activity_name = just_the_activity
        else:
            # Assume no activity or category.
            rest_after_act = act_and_rest
            expect_category = False
        return (rest_after_act, expect_category)

    # *** 2: Parse category and tags.

    def parse_cat_and_remainder(self, cat_and_remainder):
        # NOTE: cat_and_remainder may contain leading whitespace, if input
        #       was of form ``act @ cat``, not ``act@cat`` or ``act @cat``.
        # Split on any delimiter: ,|:|\n
        parts = self.re_item_sep.split(cat_and_remainder, 2)
        tags_description_sep = ''
        description_prefix = ''
        if len(parts) == 1:
            cat_and_tags = parts[0]
            unseparated_tags = None
        else:
            self.category_name = parts[0]
            cat_and_tags = None
            # parts[1] and parts[3] are the separators, e.g., ',' or ':'.
            unseparated_tags = parts[2]
            if len(parts) == 5:
                # If there's a description with a separator on same line as
                # meta data, tags_description_sep may really be description
                # prefix. So keep whitespace until we know for sure (which
                # self.re_item_sep.split accomplishes -- it puts the whitespace
                # in tags_description_sep, not the last part).
                tags_description_sep = parts[3]
                # Any whitespace between parts should be part of the former,
                # and the last part can have any of its ending whitespace
                # stripped. (Or maybe don't strip at all? Don't really care
                # if user has trailing whitespace or not, I shouldn't think!
                # We only care to remove whitespace between separators/meta.)
                #   #description_prefix = parts[4].strip()
                #   #description_prefix = parts[4].rstrip()
                description_prefix = parts[4]
            else:
                assert len(parts) == 3

        if cat_and_tags:
            cat_tags = Parser.RE_SPLIT_TAGS_AND_TAGS.split(cat_and_tags, 1)
            self.category_name = cat_tags[0]
            if len(cat_tags) == 2:
                unseparated_tags = self.hash_stamps[0] + cat_tags[1]

        self.consume_tags_and_description_prefix(
            unseparated_tags, tags_description_sep, description_prefix,
        )

    def parse_tags_and_remainder(self, tags_and_remainder):
        parts = self.re_item_sep.split(tags_and_remainder, 1)
        tags_description_sep = ''
        description_middle = ''
        if len(parts) == 3:
            # parts[1] is the sep, e.g., ',' or ':'.
            tags_description_sep = parts[1]
            description_middle = parts[2]
        self.consume_tags_and_description_prefix(
            parts[0], tags_description_sep, description_middle,
        )

    def consume_tags_and_description_prefix(
        self, unseparated_tags, tags_description_sep='', description_middle='',
    ):
        description_prefix = ''
        if unseparated_tags:
            # NOTE: re.match checks for a match only at the beginning of the string.
            match_tags = Parser.RE_SPLIT_CAT_AND_TAGS.match(unseparated_tags)
            if match_tags is not None:
                split_tags = Parser.RE_SPLIT_TAGS_AND_TAGS.split(unseparated_tags)
                self.consume_tags(split_tags)
            else:
                description_prefix = unseparated_tags
                description_prefix += tags_description_sep

        self.description = description_prefix
        self.description += description_middle
        self.description += "\n" if self.rest else ""
        self.description += self.rest

    def consume_tags(self, tags):
        tags = [tag.strip() for tag in tags]
        tags = list(filter(None, tags))
        self.tags = tags

    # ***

    def hydrate_datetimes(self):
        self.datetime1 = self.hydrate_datetime_either(
            self.datetime1, self.raw_datetime1,
        )
        self.datetime2 = self.hydrate_datetime_either(
            self.datetime2, self.raw_datetime2,
        )

    def hydrate_datetime_either(self, the_datetime, raw_datetime):
        if the_datetime or not raw_datetime:
            return the_datetime
        # Remove any trailing separator that may have been left.
        raw_datetime = self.re_item_sep.sub('', raw_datetime)
        if not the_datetime:
            the_datetime = parse_datetime_iso8601(
                raw_datetime, must=False, local_tz=self.local_tz,
            )
            if the_datetime:
                # 2018-07-02: (lb): Is this path possible?
                #   Or would we have processed ISO dates already?
                logger.warning('hydrate_datetime_either: found ISO datetime?')
        if not the_datetime:
            the_datetime = self.hydrate_datetime_friendly(
                raw_datetime, must=False,
            )
        return the_datetime

    def hydrate_datetime_friendly(self, datepart, must=False):
        settings = {
            # PREFER_DATES_FROM:    defaults to current_period.
            #                       Options are future or past.
            # SUPPORT_BEFORE_COMMON_ERA: defaults to False.
            # PREFER_DAY_OF_MONTH:  defaults to current.
            #                       Could be first and last day of month.
            # SKIP_TOKENS:          defaults to [‘t’]. Can be any string.
            # TIMEZONE:             defaults to UTC. Can be timezone abbrev
            #                       or any of tz database name as given here:
            #     https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
            # RETURN_AS_TIMEZONE_AWARE: return tz aware datetime objects in
            #                       case timezone is detected in the date string.
            # RELATIVE_BASE:        count relative date from this base date.
            #                       Should be datetime object.

            # FIXME/2018-05-22 20:46: (lb): Use RELATIVE_BASE to support
            # friendlies relative to other time, e.g.,
            #       `from 1 hour ago to 2018-05-22 20:47`

            'RETURN_AS_TIMEZONE_AWARE': False,
        }
        if self.local_tz:
            # NOTE: Uses machine-local tz, unless TIMEZONE set.
            settings['RETURN_AS_TIMEZONE_AWARE'] = True
            settings['TIMEZONE'] = self.local_tz

        parsed = dateparser.parse(datepart, settings=settings)

        if not parsed:
            parsed = None
            if must:
                raise ParserInvalidDatetimeException(_(
                    'Unable to parse datetime: {}.'
                    .format(datepart)
                ))
        elif self.skip_dateparser:
            # FIXME: Implement use of this in import routine.
            #        Caller can use RELATIVE_BASE to resolve time correctly.

            # The caller is telling us that the date is not actually
            # relative to "now". It'll call dateparser later with the
            # correct context.
            parsed = datepart

        return parsed

    # ***

    def raise_missing_datetime_one(self):
        msg = _('Expected to find a datetime.')
        raise ParserMissingDatetimeOneException(msg)

    def raise_missing_datetime_two(self):
        def _raise_missing_datetime_two():
            sep_str = comma_or_join(DATE_TO_DATE_SEPARATORS__RAW)
            msg = _(
                'Expected to find the two datetimes separated by one of: {}.'
                .format(sep_str)
            )
            raise ParserMissingDatetimeTwoException(msg)

        def comma_or_join(seq):
            # Need os.linesep or is \n fine?
            seq = [part.replace('\n', '\\n') for part in seq]
            if len(seq) > 1:
                sep_str = '{} or {}'.format(', '.join(seq[:-1]), seq[-1])
            else:
                sep_str = seq[0]
            return sep_str

        _raise_missing_datetime_two()

    def raise_missing_separator_activity(self):
        msg = _('Expected to find an "@" indicating the activity.')
        raise ParserMissingSeparatorActivity(msg)

    def raise_missing_activity(self):
        msg = _('Expected to find an Activity name.')
        raise ParserMissingActivityException(msg)


# For args, see: Parser.setup_rules().
def parse_factoid(*args, **kwargs):
    """
    Just a little shimmy-shim-shim (to Parser.dissect_raw_fact).
    """
    parser = Parser()
    err = parser.dissect_raw_fact(*args, **kwargs)
    fact_dict = {
        'start': parser.datetime1 if parser.datetime1 else None,
        'end': parser.datetime2 if parser.datetime2 else None,
        'activity': parser.activity_name.strip() if parser.activity_name else '',
        'category': parser.category_name.strip() if parser.category_name else '',
        'description': parser.description.strip() if parser.description else '',
        'tags': parser.tags if parser.tags else [],
        'warnings': parser.warnings,
    }
    return fact_dict, err




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os

import appdirs

from .singleton import Singleton

__all__ = (
    'NarkAppDirs',
)


class NarkAppDirs(appdirs.AppDirs, metaclass=Singleton):
    """Application-specific AppDirs interface.

    Note: This class is a Singleton, because its data is immutable, and it
    represents global data (that's generated from the user name and
    application details) that goes unchanged throughout the application
    lifetime. *So it's appropriate and defensible to declare this class
    as a Singleton (ALIMHO, -lb).*

    (lb): One Pythonic alternate, if you're anti the Singleton pattern, is to
    do away with the class and just have this be a simple module, i.e., expose
    user_data_dir(), etc., as module-scope functions. But that's not really
    that different, and then you end up with module-scope semi-global variables,
    anyway. And the user has to explicitly import each method, rather than just
    the class. It feels more simple, readable, and maintainable to encapsulate
    the methods in a class, and to make that class a singleton; half a dozen of
    one six of the other, if you ask me.

    For a good discussion on ways to implement Singleton in Python, and whether
    or not they're a good idea, read the classic *Creating a singleton in Python*:

        https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """

    # Although the class is a Singleton, nark needs to access the class
    # without actually creating it -- because the client (dob) needs to
    # create it (so it can set appname properly). So here's a (hacky)
    # class member to indicate when the singleton instance is prepared.
    APP_DIRS = None

    def __init__(self, *args, **kwargs):
        """Add create flag value to instance."""
        # The AppDirs takes a number of parameters:
        #   appname=None, appauthor=None, version=None,
        #   roaming=False, multipath=False,
        # but generally you just need to specify appname.
        super(NarkAppDirs, self).__init__(*args, **kwargs)

        # This constructor only called once, because Singleton.
        assert(NarkAppDirs.APP_DIRS is None)
        # A hacky class variable so nark knows when the instance is ready.
        NarkAppDirs.APP_DIRS = self

        # FIXME: (lb): I'm not super cool with this side-effect:
        #          Calling any property herein will cause its
        #          directory path to be created! Creating paths
        #          should be a deliberate action and not a side effect
        #          of just asking for a path. In any case, it currently
        #          works this way, so just rolling with the flow, for now.
        #        See Click: it has concept of lazy-creating paths, i.e.,
        #          only create path when a file therein opened for write.
        self.create = True

    @property
    def user_data_dir(self):
        """Return ``user_data_dir``."""
        directory = appdirs.user_data_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            roaming=self.roaming,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def site_data_dir(self):
        """Return ``site_data_dir``."""
        directory = appdirs.site_data_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            multipath=self.multipath,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def user_config_dir(self):
        """Return ``user_config_dir``."""
        directory = appdirs.user_config_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            roaming=self.roaming,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def site_config_dir(self):
        """Return ``site_config_dir``."""
        directory = appdirs.site_config_dir(
            self.appname,
            self.appauthor,
            version=self.version,
            multipath=self.multipath,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def user_cache_dir(self):
        """Return ``user_cache_dir``."""
        directory = appdirs.user_cache_dir(
            self.appname, self.appauthor, version=self.version,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    @property
    def user_log_dir(self):
        """Return ``user_log_dir``."""
        directory = appdirs.user_log_dir(
            self.appname, self.appauthor, version=self.version,
        )
        if self.create:
            self._ensure_directory_exists(directory)
        return directory

    def _ensure_directory_exists(self, directory):
        """Ensure that the passed path exists."""
        if not os.path.lexists(directory):
            os.makedirs(directory)
        return directory




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from . import BaseManager
from ..items.tag import Tag


class BaseTagManager(BaseManager):
    """
    Base class defining the minimal API for a TagManager implementation.
    """

    def __init__(self, *args, **kwargs):
        super(BaseTagManager, self).__init__(*args, **kwargs)

    # ***

    def save(self, tag):
        """
        Save a Tag to the backend.

        Args:
            tag (nark.Tag): Tag instance to be saved.

        Returns:
            nark.Tag: Saved Tag

        Raises:
            TypeError: If the ``tag`` parameter is not a valid ``Tag`` instance.
        """
        return super(BaseTagManager, self).save(tag, Tag, named=True)

    # ***

    def get_or_create(self, tag):
        """
        Check if we already got a tag with that name, if not create one.

        This is a convenience method as it seems sensible to rather implement
        this once in our controller than having every client implementation
        deal with it anew.

        It is worth noting that the lookup completely ignores any PK contained in the
        passed tag. This makes this suitable to just create the desired Tag
        and pass it along. One way or the other one will end up with a persisted
        db-backed version.

        Args:
            tag (nark.Tag or None): The categories.

        Returns:
            nark.Tag or None: The retrieved or created tag. Either way,
                the returned Tag will contain all data from the backend, including
                its primary key.
        """
        self.store.logger.debug(_("'{}' has been received.'.".format(tag)))
        if tag:
            try:
                tag = self.get_by_name(tag)
            except KeyError:
                tag = Tag(tag)
                tag = self._add(tag)
        else:
            # We want to allow passing ``tag=None``, so we normalize here.
            tag = None
        return tag

    # ***

    def _add(self, tag):
        """
        Add a ``Tag`` to our backend.

        Args:
            tag (nark.Tag): ``Tag`` to be added.

        Returns:
            nark.Tag: Newly created ``Tag`` instance.

        Raises:
            ValueError: When the tag name was already present! It is supposed to be
            unique.
            ValueError: If tag passed already got an PK. Indicating that update would
                be more appropriate.
        """
        raise NotImplementedError

    # ***

    def _update(self, tag):
        """
        Update a ``Tags`` values in our backend.

        Args:
            tag (nark.Tag): Tag to be updated.

        Returns:
            nark.Tag: The updated Tag.

        Raises:
            KeyError: If the ``Tag`` can not be found by the backend.
            ValueError: If the ``Tag().name`` is already being used by
                another ``Tag`` instance.
            ValueError: If tag passed does not have a PK.
        """
        raise NotImplementedError

    # ***

    def remove(self, tag):
        """
        Remove a tag.

        Any ``Fact`` referencing the passed tag will have this tag removed.

        Args:
            tag (nark.Tag): Tag to be updated.

        Returns:
            None: If everything went ok.

        Raises:
            KeyError: If the ``Tag`` can not be found by the backend.
            TypeError: If tag passed is not an nark.Tag instance.
            ValueError: If tag passed does not have an pk.
        """
        raise NotImplementedError

    # ***

    def get(self, pk):
        """
        Get an ``Tag`` by its primary key.

        Args:
            pk (int): Primary key of the ``Tag`` to be fetched.

        Returns:
            nark.Tag: ``Tag`` with given primary key.

        Raises:
            KeyError: If no ``Tag`` with this primary key can be found by the backend.
        """

        raise NotImplementedError

    # ***

    def get_by_name(self, name):
        """
        Look up a tag by its name.

        Args:
            name (str): Unique name of the ``Tag`` to we want to fetch.

        Returns:
            nark.Tag: ``Tag`` with given name.

        Raises:
            KeyError: If no ``Tag`` with this name was found by the backend.
        """
        raise NotImplementedError

    # ***

    def get_all(
        self,
        include_usage=True,
        count_results=False,
        since=None,
        until=None,
        deleted=False,
        hidden=False,
        key=None,
        search_term='',
        activity=False,
        category=False,
        sort_col='',
        sort_order='',
        limit='',
        offset='',
    ):
        """
        Get all tags, with filtering and sorting options.

        Returns:
            list: List of all Tags present in the database,
                  ordered by lower(name), or most recently
                  used; possibly filtered by a search term.
        """
        raise NotImplementedError




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import logging
import sys

from gettext import gettext as _

__all__ = (
    'must_verify_log_level',
    'get_log_level_safe',
    'get_log_name_safe',
    # Private:
    #  'LOG_LEVELS',
)


this = sys.modules[__name__]


# ***
# *** Config function: log level helpers.
# ***

# Subset (and lowercase) of logging._nameToLevel.
this.LOG_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL,
}


# MEH/2019-01-17: Deal with this when refactoring config:
#   If cli_log_level is wrong, app just logs a warning.
#   But for some reason, here, if sql_log_level is wrong,
#   app dies. Should probably just warning instead and
#   get on with life... print colorful stderr message,
#   but live.
#     See also: nark/nark/control.py and nark/nark/helpers/logging.py
#     have log_level functions, should probably consolidate this!
def must_verify_log_level(level_name):
    if isinstance(level_name, int):
        return level_name
    try:
        log_level = this.LOG_LEVELS[level_name.lower()]
    except AttributeError:
        msg = _(
            " (Unrecognized log level type in config: “{}”. Try a string from: {}.)"
        ).format(level_name, ', '.join(this.LOG_LEVELS))
        raise ValueError(msg)
    except KeyError:
        msg = _(
            " (Unrecognized log level value in config: “{}”. Try one of: {}.)"
        ).format(level_name, ', '.join(this.LOG_LEVELS))
        raise ValueError(msg)
    return log_level


def get_log_level_safe(level_name):
    try:
        log_level = must_verify_log_level(level_name)
    except ValueError:
        # MAYBE: (lb): Complain to user that their config value is bad.
        log_level = logging.WARNING

    # (lb): A wee bit of a hack! Don't log during the dob-complete
    #   command, lest yuck!
    # MEH/2020-01-29: If logging to file, don't change level.
    if (len(sys.argv) == 2) and (sys.argv[1] == 'complete'):
        # Disable for dob-complete.
        return logging.CRITICAL + 1
    return log_level


def get_log_name_safe(level):
    return logging.getLevelName(level)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""nark provides generic time tracking functionality."""

import os
import time

__all__ = (
    'get_version',
    '__package_name__',
    '__time_0__',
    '__PROFILING__',
)

__PROFILING__ = True
# DEVS: Comment this out to see load times summary.
__PROFILING__ = False
__time_0__ = time.time()

# (lb): Seems a little redundant (see setup.cfg:[metadata]name)
# but not sure if way to get programmatically. This is closest
# solution that avoids hardcoding the library name in strings
# (which is something linter or runtime won't catch if wrong).
__package_name__ = 'nark'


# MAYBE: (lb): This feels like a package one-off... which is exactly
# what we need, another PyPI package to manage....
def get_version(package_name=None, reference_file=None, include_head=False):
    """Returns the installed package version, or '<none>'.

    In lieu of always setting __version__ -- and always loading pkg_resources --
    use a method to avoid incurring startup costs if the version is not needed.
    """
    def resolve_vers():
        dist_version = version_installed()
        if include_head:
            repo_version = version_from_repo()
            if repo_version:
                dist_version = '{} ({})'.format(dist_version, repo_version)
        return dist_version

    def version_installed():
        # - This returns the version most recently pip-installed. That is, if
        #   you install local sources and have committed code but not run the
        #   pip-install again, this shows the older version.
        from pkg_resources import get_distribution, DistributionNotFound
        try:
            distrib_name = package_name or __package_name__
            return get_distribution(distrib_name).version
        except DistributionNotFound:
            # This would be really weird, no?
            return '<none!?>'

    def version_from_repo():
        try:
            return version_from_tags()
        # Note: ModuleNotFoundError in Py3.6+, so using less specific ImportError.
        except ImportError:
            # No setuptools_scm package installed.
            return ''
        except LookupError:
            # Path containing .git/ not a repo after all.
            return '<none?!>'

    def version_from_tags():
        # Try to get the version from SCM. Obvi, this is intended for devs,
        # as normal users will likely not have setuptools_scm installed.
        import setuptools_scm
        # For whatever reason, relative_to does not work, (lb) thought it would.
        #   return setuptools_scm.get_version(relative_to=__file__)
        # So figure out the root path of the repo. In lieu of something robust,
        # like `git rev-parse --show-toplevel`, look for '.git/' ourselves.
        cur_path = reference_file or __file__
        while cur_path and cur_path != os.path.dirname(cur_path):
            cur_path = os.path.dirname(cur_path)
            proj_git = os.path.join(cur_path, '.git')
            if os.path.exists(proj_git):
                # Get version from setuptools_scm, and git tags.
                # This is similar to a developer running, e.g.,
                #   python setup.py --version
                return setuptools_scm.get_version(root=cur_path)
        # Not .git/ found. Package probably installed to site-packages/.
        return ''

    return resolve_vers()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from collections import namedtuple
from datetime import datetime
from operator import attrgetter

from ..helpers import fact_time, format_time
from ..helpers.format_text import format_value_truncate
from ..helpers.parsing import parse_factoid
from .activity import Activity
from .category import Category
from .item_base import BaseItem
from .tag import Tag

try:
    from math import inf
except ImportError:
    # < Python 3.5.
    inf = float('inf')


__all__ = (
    'SinceTimeBegan',
    'UntilTimeStops',
    'FactTuple',
    'Fact',
)


SinceTimeBegan = datetime(1, 1, 1)


# FIXME/2020-01-27: #Y10Kbug.
UntilTimeStops = datetime(9999, 12, 31, 23, 59, 59)


FactTuple = namedtuple(
    'FactTuple',
    (
        'pk',
        'activity',
        'start',
        'end',
        'description',
        'tags',
        'deleted',
        'split_from',
    ),
)


class Fact(BaseItem):
    """Storage agnostic class for facts."""
    def __init__(
        self,
        activity,
        start,
        end=None,
        pk=None,
        description=None,
        tags=None,
        deleted=False,
        split_from=None,
    ):
        """
        Initiate our new instance.

        Args:
            activity (nark.Activity): Activity associated with this fact.

            start (datetime.datetime): Start datetime of this fact.

            end (datetime.datetime, optional): End datetime of this fact.
                Defaults to ``None``.

            pk (optional): Primary key used by the backend to identify this instance.
                Defaults to ``None``.

            description (str, optional): Additional information relevant to this
                singular fact. Defaults to ``None``.

            tags (Iterable, optional): Iterable of ``strings`` identifying *tags*.
                Defaults to ``None``.

            deleted (bool, optional): True if fact was deleted/edited/split.

            split_from (nark.Fact.id, optional): ID of deleted fact this
                fact succeeds.
        """
        super(Fact, self).__init__(pk, name=None)
        assert activity is None or isinstance(activity, Activity)
        self.activity = activity
        self.start = start
        self.end = end
        self.description = description

        self.tags = []
        self.tags_replace(tags)

        # (lb): Legacy Hamster did not really have an edit-fact feature.
        # Rather, when the user "edited" a Fact, Hamster would delete
        # the existing row and make a new one. This is very unwikilike!
        # To preserve history, let's instead mark edited Facts deleted.
        # FIXME/2018-05-23 10:56: (lb): Add column to record new Facts ID?
        self.deleted = bool(deleted)

        self.split_from = split_from

    def __eq__(self, other):
        if other is not None and not isinstance(other, FactTuple):
            other = other.as_tuple()

        return self.as_tuple() == other

    def equal_sans_end(self, other):
        if other is not None and not isinstance(other, FactTuple):
            other = other.as_tuple(sans_end=True)

        return self.as_tuple(sans_end=True) == other

    def __hash__(self):
        """Naive hashing method."""
        return hash(self.as_tuple())

    def __gt__(self, other):
        return self.sorty_tuple > other.sorty_tuple

    def __lt__(self, other):
        return self.sorty_tuple < other.sorty_tuple

    @property
    def sorty_times(self):
        fact_end = self.end if self.end is not None else UntilTimeStops
        return (self.start, fact_end)

    @property
    def sorty_tuple(self):
        fact_end = self.end if self.end is not None else UntilTimeStops
        fact_pk = self.pk if self.pk is not None else -inf
        return (self.start, fact_end, fact_pk)

    def __str__(self):
        return self.friendly_str()

    def __repr__(self):
        return self.as_kvals()

    def as_kvals(self):
        parts = []
        for key in sorted(self.__dict__.keys()):
            if key == 'name':
                assert self.name is None
                continue  # Part of BaseItem, and not used by Fact.
            elif key == 'tags':
                val = repr(self.tags_sorted)
            else:
                # Not that nark knows anything about this, but dob's derived
                # class, FactDressed, has a next_fact and prev_fact pointer.
                # To avoid a deadly RecursionError by following pointers of a
                # doubly-linked list, check that the attribute value is not
                # another Fact. (For the sake of coupling, FactDressed could
                # just supply its own __repr__() function, but this function
                # is already doing a lot, so might as well toss in the kitchen
                # sink.)
                raw = getattr(self, key)
                if isinstance(raw, self.__class__):
                    val = id(raw)
                else:
                    val = repr(raw)
            parts.append(
                "{key}={val}".format(key=key, val=val)
            )
        repred = "Fact({})".format(', '.join(parts))
        return repred

    def as_tuple(self, include_pk=True, sans_end=False):
        """
        Provide a tuple representation of this facts relevant attributes.

        Args:
            include_pk (bool): Whether to include the instances pk or not. Note that if
            ``False`` ``tuple.pk = False``!

        Returns:
            nark.FactTuple: Representing this categories values.
        """
        pk = self.pk
        if not include_pk:
            pk = False

        activity_tup = self.activity and self.activity.as_tuple(include_pk=include_pk)

        sorted_tags = self.tags_sorted
        ordered_tags = [tag.as_tuple(include_pk=include_pk) for tag in sorted_tags]

        end_time = -1 if sans_end else self.end

        return FactTuple(
            pk=pk,
            activity=activity_tup,
            start=self.start,
            end=end_time,
            description=self.description,
            tags=frozenset(ordered_tags),
            deleted=bool(self.deleted),
            split_from=self.split_from,
        )

    def copy(self, include_pk=True):
        """
        """
        new_fact = self.__class__(
            activity=self.activity,
            start=self.start,
            end=self.end,
            description=self.description,
            # self.tags might be an sqlalchemy.orm.collections.InstrumentedList
            # and calling list() on it will create a new list of what could be
            # nark.backends.sqlalchemy.objects.AlchemyTag.
            tags=list(self.tags),
            deleted=bool(self.deleted),
            split_from=self.split_from,
        )
        if include_pk:
            new_fact.pk = self.pk
        return new_fact

    def equal_fields(self, other):
        """
        Compare this instances fields with another fact. This excludes comparing the PK.

        Args:
            other (Fact): Fact to compare this instance with.

        Returns:
            bool: ``True`` if all fields but ``pk`` are equal, ``False`` if not.

        Note:
            This is particularly useful if you want to compare a new ``Fact`` instance
            with a freshly created backend instance. As the latter will probably have a
            primary key assigned now and so ``__eq__`` would fail.
        """
        return self.as_tuple(include_pk=False) == other.as_tuple(include_pk=False)

    # ***

    LOCALIZE = False

    @classmethod
    def localize(cls, localize=None):
        if localize is not None:
            cls.LOCALIZE = localize
        return cls.LOCALIZE

    # ***

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        """
        Make sure that we receive a ``datetime.datetime`` instance,
          or relative time string.

        Args:
            start (datetime.datetime, str): Start datetime of this ``Fact``.

        Raises:
            TypeError: If we receive something other than a ``datetime.datetime``
                (sub-)class or ``None``.
        """
        # MOTE: (lb): The AlchemyFact class derives from this class, but it
        # has columns of the same names as theses @property definitions, e.g.,
        # `start`. So when you set, e.g, `self.start = X` from the AlchemyFact
        # class, it does not call this base class' @setter for start. So don't
        # use self._start except in self.start()/=.
        self._start = fact_time.must_be_datetime_or_relative(start)

    @property
    def start_fmt_utc(self):
        """FIXME: Document"""
        if not self.start:
            return ''
        # Format like: '%Y-%m-%d %H:%M:%S%z'
        return format_time.isoformat_tzinfo(self.start, sep=' ', timespec='seconds')

    @property
    def start_fmt_local(self):
        """FIXME: Document"""
        if not self.start:
            return ''
        # Format like: '%Y-%m-%d %H:%M:%S'
        return format_time.isoformat_tzless(self.start, sep=' ', timespec='seconds')

    # ***

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        """
        Make sure that we receive a ``datetime.datetime`` instance.

        Args:
            end (datetime.datetime): End datetime of this ``Fact``.

        Raises:
            TypeError: If we receive something other than a ``datetime.datetime``
                (sub-)class or ``None``.
        """
        self._end = fact_time.must_be_datetime_or_relative(end)

    @property
    def end_fmt_utc(self):
        """FIXME: Document"""
        if not self.end:
            return ''
        return format_time.isoformat_tzinfo(self.end, sep=' ', timespec='seconds')

    @property
    def end_fmt_local(self):
        """FIXME: Document"""
        if not self.end:
            return ''
        return format_time.isoformat_tzless(self.end, sep=' ', timespec='seconds')

    @property
    def end_fmt_local_nowwed(self):
        """FIXME: Document"""
        if not self.end:
            return '{} <now>'.format(self.end_fmt_local_or_now)
        return self.end_fmt_local

    @property
    def end_fmt_local_or_now(self):
        if not self.end:
            return '{}'.format(format_time.isoformat_tzless(
                self.time_now, sep=' ', timespec='seconds',
            ))
        return self.end_fmt_local

    # ***

    @property
    def momentaneous(self):
        if self.times_ok and self.start == self.end:
            return True
        return False

    @property
    def time_now(self):
        return datetime.now() if self.localize else datetime.utcnow()

    @property
    def times(self):
        return (self.start, self.end)

    @property
    def times_ok(self):
        if isinstance(self.start, datetime) and isinstance(self.end, datetime):
            return True
        return False

    # ***

    def delta(self):
        """
        Provide the offset of start to end for this fact.

        Returns:
            datetime.timedelta or None: Difference between start- and end datetime.
                If we only got a start datetime, return ``None``.
        """
        end_time = self.end
        if not end_time:
            end_time = self.time_now

        return end_time - self.start

    def format_delta(self, style='%M'):
        """
        Return a string representation of ``Fact().delta``.

        Args:
            formatting (str): Specifies the output format.

              Valid choices are:
                * ``'%M'``: As minutes, rounded down.
                * ``'%H:%M'``: As 'hours:minutes'. rounded down.
                * ``'HHhMMm'``: As '{hours} hour(s) {minutes} minute(s)'.
                * ``''``: As human friendly time.

        Returns:
            str: Formatted string representing this fact's *duration*.
        """
        return format_time.format_delta(self.delta(), style=style)

    # ***

    @property
    def midpoint(self):
        if not self.times_ok:
            return None
        midpoint = self.end - ((self.end - self.start) / 2)
        return midpoint

    @property
    def time_of_day_midpoint(self):
        if not self.midpoint:
            return ''
        clock_sep = ' ◐ '
        hamned = '{0}'.format(
            self.midpoint.strftime("%a %d %b %Y{0}%I:%M %p").format(clock_sep),
            # FIXME: (lb): Add Colloquial TOD suffix, e.g., "morning".
        )
        return hamned

    def time_of_day_humanize(self, show_now=False):
        if not self.times_ok and not show_now:
            return ''
        clock_sep = ' ◐ '
        wkd_day_mon_year = self.start.strftime("%a %d %b %Y")
        text = self.start.strftime("{0}{1}%I:%M %p").format(
            wkd_day_mon_year, clock_sep,
        )
        if self.end == self.start:
            return text
        text += _(" — ")
        end_time = self.end if self.end is not None else self.time_now
        text += end_time.strftime("%I:%M %p")
        end_wkd_day_mon_year = end_time.strftime("%a %d %b %Y")
        if end_wkd_day_mon_year == wkd_day_mon_year:
            return text
        text += " "
        text += end_wkd_day_mon_year
        return text

    # ***

    @property
    def activity_name(self):
        """..."""
        try:
            return self.activity.name
        except AttributeError:
            return ''

    # +++

    @property
    def category(self):
        """Just a convenience shim to underlying category object."""
        return self.activity.category

    @property
    def category_name(self):
        """..."""
        try:
            return self.activity.category.name
        except AttributeError:
            return ''

    # +++

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        """"
        Normalize all descriptions that evaluate to ``False``.
        Store everything else as string.
        """
        if description:
            description = str(description)
        else:
            description = None
        self._description = description

    # +++

    def tagnames(self, *args, **kwargs):
        return self.oid_tags(*args, **kwargs)

    def tagnames_sorted_formatted(self, format_tagname):
        return [format_tagname(tag) for tag in self.tags_sorted]

    def tags_replace(self, tags):
        new_tags = set()
        for tagn in set(tags) if tags else set():
            if isinstance(tagn, Tag):
                tag = tagn
            else:
                # Retain existing tags rather than replacing with new Tag()s.
                # This is useful for clients that compare changes to a Fact,
                # to avoid a Fact appearing edited when not. For instance, in
                # clyde, after the user uses the Awesome Prompt, even if they
                # just open it and quit it, the returned tags is a list of strs.
                # And clyde sends that strings list to this tags_replace method.
                # So be sure to look for existing tag matches and use those --
                # if we built new Tag()s instead, the new tags would have None
                # for the pk, rather than the ID originally read from the store,
                # and then the fact would false-positive look edited (dirty).
                # And then clyde would bug you to save your unedited Fact, etc.
                tag = next(
                    (tag for tag in self.tags if tag.name == tagn),
                    Tag(name=tagn),
                )
            new_tags.add(tag)
        # (lb): Do this in one swoop, and be sure to assign a list; when wrapped
        # by SQLAlchemy, if tried to set to, e.g., set(), complains:
        #   TypeError: Incompatible collection type: set is not list-like
        # (That error is from orm.attribute.CollectionAttributeImpl.set.)
        self.tags = list(new_tags)

    @property
    def tags_sorted(self):
        return sorted(list(self.tags), key=attrgetter('name'))

    # ***

    # ’oid, as in Factoid: these methods help make the friendly, parsable Factoid.

    def oid_colorize(self, oid_part, oid_text):
        """Returns stylized text for a specific part of the Fact*oid*.

        Clients that display Facts where ornamentation matters can override
        this method. This base class implementation is a no-op.
        """
        return oid_text

    def oid_actegory(self, shellify=False, omit_empty_actegory=False):
        # (lb): We can skip delimiter after time when using ISO 8601.
        if not self.activity_name:
            if not self.category_name:
                act_cat = '' if omit_empty_actegory else '@'
            else:
                act_cat = '@{}'.format(self.category_name)
        else:
            act_cat = (
                '{}@{}'.format(
                    self.activity_name,
                    self.category_name,
                )
            )
        act_cat = self.oid_colorize('act@gory', act_cat)
        act_cat = '"{}"'.format(act_cat) if act_cat and shellify else act_cat
        return act_cat

    def oid_description(self, cut_width=None, sep=', '):
        description = self.description or ''
        if description:
            if cut_width is not None:
                description = format_value_truncate(description, cut_width)
            description = '{}{}'.format(sep, description)
        return description

    # (lb): People associate tags with pound signs -- like, #hashtag!
    # But Bash, and other shells, use octothorpes to start comments.
    # The user can tell Bash to interpret a pound sign as input by
    # "#quoting" it, or by \#delimiting it. Hamster also lets the user
    # use an '@' at symbol instead (not to be confused with typical
    # social media usage of '@' to refer to other users or people).
    # By default, this function assumes the tags do not need special
    # delimiting, and that the pound sign is fine.
    def oid_tags(
        self,
        hashtag_token='#',
        quote_tokens=False,
    ):
        def format_tagname(tag):
            tagged = '{}{}'.format(
                self.oid_colorize('#', hashtag_token),
                self.oid_colorize('tag', tag.name),
            )
            tagged = self.oid_colorize('#tag', tagged)
            if quote_tokens:
                tagged = '"{}"'.format(tagged)
            return tagged

        # NOTE: The returned string includes leading space if nonempty!
        tagnames = ''
        if self.tags:
            tagnames = ' '.join(self.tagnames_sorted_formatted(format_tagname))
        return tagnames

    # +++

    def friendly_str(
        self,
        shellify=False,
        description_sep=': ',
        tags_sep=': ',
        localize=False,
        include_id=False,
        cut_width=None,
        show_elapsed=False,
        omit_empty_actegory=False,
    ):
        """
        Flexible Fact serializer.
        """
        def _friendly_str():
            # MAYBE/2019-01-28: Truncate meta per cut_width or similar.
            meta = assemble_parts()
            result = format_result(meta)
            return result

        def assemble_parts():
            parts = [
                get_id_string(),
                get_times_string(),
                self.oid_actegory(shellify, omit_empty_actegory),
            ]
            parts_str = ' '.join(list(filter(None, parts)))
            tags = get_tags_string()
            parts_str += tags_sep + tags if tags else ''
            parts_str += _(" [del]") if self.deleted else ''
            return parts_str

        def format_result(meta):
            result = '{fact_meta}{description}'.format(
                fact_meta=meta,
                description=self.oid_description(cut_width, description_sep),
            )
            return result

        def get_id_string():
            if not include_id:
                return ''
            # Format the 🏭 🆔 width to be consistent. Assume lifetime of facts?
            # [ [fact]ory ↑ ↑ ID ]
            # - 6 digits: 999,999 facts over 100 years would be ~27 facts per day.
            return self.oid_colorize(
                'pk',
                '(🏭 {})'.format(self.pk and '{:6d}'.format(self.pk) or 'None'),
            )

        def get_times_string():
            times = ''
            times += get_times_string_start()
            times += get_times_string_end(times)
            times += get_times_duration()
            return times

        def get_times_string_start():
            if not self.start:
                return ''
            if not self.localize:
                start_time = self.start_fmt_utc
            else:
                start_time = self.start_fmt_local
            start_time = self.oid_colorize('start', start_time)
            return start_time

        def get_times_string_end(times):
            # NOTE: The CLI's DATE_TO_DATE_SEPARATORS[0] is 'to'.
            prefix = self.oid_colorize('to', ' to ') if times else ''
            if not self.end:
                if not times:
                    end_time = ''
                else:
                    # (lb): What's a good term here? '<ongoing>'? Or just 'now'?
                    end_time = _('<now>')
            elif not self.localize:
                end_time = self.end_fmt_utc
            else:
                end_time = self.end_fmt_local
            end_time = self.oid_colorize('end', end_time)
            return prefix + end_time

        def get_times_duration():
            if not show_elapsed:
                return ''
            duration = ' [{}]'.format(self.format_delta(style=''))
            return self.oid_colorize('duration', duration)

        def get_tags_string():
            # (lb): There are three ways to "shellify" a hashtag token:
            #         1.) "#quote" it;
            #         2.) \#delimit it; or
            #         3.) use the inoffensive @ symbol instead of #.
            # Let's do 1.) by default, because most people associate the pound
            # sign with tags, because quotes are less offensive than a slash,
            # and because the @ symbol makes me think of "at'ing someone".
            #   Nope:  hashtag_token = '@' if shellify else '#'
            return self.oid_tags(quote_tokens=shellify)

        # ***

        return _friendly_str()

    # +++

    def get_serialized_string(self, shellify=False):
        """
        Return a canonical, "stringified" version of the Fact.

        - Akin to: encoding/flattening/marshalling/packing/pickling.

        This function is mostly meant for machines, not for people.

        - Generally, use ``__str__`` if you want a human-readable string.

          I.e., one whose datetimes are localized relative to the Fact.
          This serializing function defaults to using UTC.

        - Use this function to encode a Fact in a canonical way, which can
          be consumed again later, i.e., using ``Fact.create_from_factoid``.

        - A complete serialized fact might look like this:

              2016-02-01 17:30 to 2016-02-01 18:10 making plans@world domination
              #tag 1 #tag 2, description

          - Note that nark is very unassuming with whitespace. It can be
            used in the Activity and Category names, as well as in tags.

        Attention:

            ``Fact.tags`` is a set and hence unordered. In order to provide
            a deterministic canonical return string, we sort tags by name.
            This is purely cosmetic and does not imply any actual ordering
            of those facts on the instance level.

        Returns:
            str: Canonical string encoding all available fact info.
        """
        return self.friendly_str(shellify=shellify)

    # +++

    @property
    def short(self):
        """
        A brief Fact one-liner.

        (lb): Not actually called by any code, but useful for debugging!
        """
        # HARDCODED: Truncate description at 39 chars. If this method was
        # actually used, and not just be a DEV on the PDB prompt, we might
        # care to move this value to the config, or at least a class attr.
        return self.friendly_str(include_id=True, cut_width=39)

    # ***

    @classmethod
    def create_from_factoid(
        cls,
        factoid,
        time_hint='verify_none',
        separators=None,
        lenient=False,
    ):
        """
        Construct a new ``nark.Fact`` from a string of fact details,
            or factoid.

        NOTE: This naïvely creates a new Fact and does not check against
        other Facts for integrity. It's up to the caller to see if the new
        Fact conflicts with existing Facts presently in the system.

        Args:
            factoid (str): Raw fact to be parsed.

            time_hint (str, optional): One of:
                'verify_none': Do not expect to find any time encoded in factoid.
                'verify_both': Expect to find both start and end times.
                'verify_start': Expect to find just one time, which is the start.
                'verify_end': Expect to find just one time, which is the end.
                'verify_then': Time specifies new start; and back-fill interval gap.
                'verify_after': No time spec. Start new Fact at time of previous end.

            lenient (bool, optional): If False, parser raises errors on misssing
                mandatory components (such as time or activity). (Category,
                tags, and description are optional.)

        Returns:
            nark.Fact: New ``Fact`` object constructed from factoid.

        Raises:
            ValueError: If we fail to extract at least ``start`` or ``activity.name``.
            ValueError: If ``end <= start``.
            ParserException: On parser error, one of the many ParserException
                derived classes will be raised.
        """
        parsed_fact, err = parse_factoid(
            factoid,
            time_hint=time_hint,
            separators=separators,
            lenient=lenient,
        )

        new_fact = cls.create_from_parsed_fact(
            parsed_fact, lenient=lenient,
        )

        return new_fact, err

    @classmethod
    def create_from_parsed_fact(
        cls,
        parsed_fact,
        lenient=False,
        **kwargs
    ):
        start = parsed_fact['start']
        end = parsed_fact['end']
        # Verify that start > end, if neither are None or not a datetime.
        start, end = fact_time.must_not_start_after_end((start, end))

        activity = ''
        activity_name = parsed_fact['activity']
        if activity_name:
            activity = Activity(activity_name)
        elif lenient:
            activity = Activity(name='')
        else:
            raise ValueError(_('Unable to extract activity name'))

        category_name = parsed_fact['category']
        if category_name:
            activity.category = Category(category_name)

        description = parsed_fact['description']

        tags = parsed_fact['tags']

        return cls(
            activity,
            start,
            end=end,
            description=description,
            tags=tags,
            **kwargs
        )




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures that are of general use."""

from configobj import ConfigObj

import pytest

from nark.config.log_levels import LOG_LEVELS


@pytest.fixture(params=list(LOG_LEVELS.keys()) + [123, ])
def log_level_valid_parametrized(request):
    """Return each of the valid log level strings."""
    return request.param


@pytest.fixture(params=(None, '123', 'abc', ''))
def log_level_invalid_parametrized(request):
    """Return selection of invalid log level strings."""
    return request.param


@pytest.fixture
def configobj_instance(request):
    """Provide a ``ConfigObj`` instance and its expected config dict."""

    config = ConfigObj()
    config['db'] = {}
    config['db']['orm'] = 'sqlalchemy'
    config['db']['engine'] = 'sqlite'
    config['db']['path'] = '/tmp/hamster.db'
    config['db']['host'] = 'www.example.com'
    config['db']['port'] = 22
    config['db']['name'] = 'hamster'
    config['db']['user'] = 'hamster'
    config['db']['password'] = 'hamster'
    config['dev'] = {}
    config['dev']['lib_log_level'] = 'WARNING'
    config['dev']['sql_log_level'] = 'debug'
    config['time'] = {}
    config['time']['allow_momentaneous'] = False
    config['time']['day_start'] = '05:00:00'
    config['time']['fact_min_delta'] = 60

    expectation = {
        'db': {
            'orm': 'sqlalchemy',
            'engine': 'sqlite',
            'path': '/tmp/hamster.db',
            'host': 'www.example.com',
            'port': '22',
            'name': 'hamster',
            'user': 'hamster',
            'password': 'hamster',
        },
        'dev': {
            'lib_log_level': 'WARNING',
            'sql_log_level': 'debug',
        },
        'time': {
            'allow_momentaneous': 'False',
            'day_start': '05:00:00',
            'fact_min_delta': '60',
            # MAYBE: (lb): Consider fiddling with day_start and fact_min_delta
            # in specific tests and leaving them set to factory defaults here:
            #   'day_start': '',
            #   'fact_min_delta': 0,
        },
    }

    return config, expectation




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides several time-related convenience functions."""

from gettext import gettext as _

import datetime
import re


__all__ = (
    'datetime_from_clock_prior',
    'datetime_from_clock_after',
    'day_end_datetime',
    'day_end_time',
    'must_be_datetime_or_relative',
    'must_not_start_after_end',
    'RE_PATTERN_RELATIVE_CLOCK',
    'RE_PATTERN_RELATIVE_DELTA',
)


# ***

def datetime_from_clock_prior(dt_relative, clock_time):
    # FIXME/MEH/2018-05-21 11:32: (lb): I'm guessing this doesn't work
    # across the "fold", e.g., 2 AM on daylight savings "fall back"
    # occurs twice, and in Python, the first time, fold=0, and the
    # second time, fold=1.
    new_dt = dt_relative.replace(
        hour=int(clock_time[0]),
        minute=int(clock_time[1]),
        second=int(clock_time[2]),
    )
    if new_dt > dt_relative:
        new_dt -= datetime.timedelta(days=1)
    return new_dt


def datetime_from_clock_after(dt_relative, clock_time):
    # FIXME/MEH/2018-05-21 11:32: (lb): Ignoring so-called "fold"/DST issue.
    new_dt = dt_relative.replace(
        hour=int(clock_time[0]),
        minute=int(clock_time[1]),
        second=int(clock_time[2]),
    )
    if new_dt < dt_relative:
        new_dt += datetime.timedelta(days=1)
    return new_dt


# ***

def day_end_datetime(end_date, start_time=None):
    """
    Convert a given end date to its proper datetime, based on a day start time.

    Args:
        end_date (datetime.date): Raw end date that is to be adjusted.
        start_time (string): Clock time of start of day.

    Returns:
        datetime.datetime: The adjusted end datetime for a given date,
          based on a specific start_time.

    Example:
        Given a ``start_time`` of ``5:30`` and an end date of ``2015-04-01``,
          the active end datetime is ``2015-04-02 5:29``, to account for an
          actual start datettime of ``2015-04-01 5:30``. The gist is that a
          *work day* does not match a *calendar* (24-hour) day; it depends on
          what the user considers their "daily start time". (Though for many
          users, they'll leave day_start untouched, in which case a single
          day covers a normal 24-hour span, i.e., from 00:00 to 23:59.

    Note:
        An alternative implementation can be found in legacy hamster:
          ``hamster.storage.db.Storage.__get_todays_facts``.
    """
    start_time = start_time or datetime.time(0, 0, 0)
    end_time = day_end_time(start_time)
    if start_time == datetime.time(0, 0, 0):
        # The start time is midnight, so the end time is 23:59:59
        # on the same day.
        assert end_time == datetime.time(23, 59, 59)
        end = datetime.datetime.combine(end_date, end_time)
    else:
        # The start time is not midnight, so the end time is
        # on the following day.
        end = datetime.datetime.combine(end_date, end_time)
        end += datetime.timedelta(days=1)
    return end


def day_end_time(start_time):
    """
    Get the day end time given the day start. This assumes full 24h day.

    Args:
        start_time (string): Clock time of start of day.
    """
    # NOTE: Because we're only returning the time, we don't need the
    #       static "now" from the controller.
    today_date = datetime.date.today()
    start_datetime = datetime.datetime.combine(today_date, start_time)
    end_datetime = start_datetime - datetime.timedelta(seconds=1)
    end_time = end_datetime.time()
    return end_time


# ***

# (lb) See comment atop pattern_date in parse_time about allowing
#   \d{4} (without :colon). Here's the stricter pattern:
#    '^(?P<hours>\d{2}):(?P<minutes>\d{2})$'
RE_PATTERN_RELATIVE_CLOCK = re.compile(
    '^(?P<hours>\d{1,2}):?(?P<minutes>\d{2})(:(?P<seconds>\d{2}))?$'
)


RE_PATTERN_RELATIVE_DELTA = re.compile(
    '^(?P<signage>[-+])?((?P<hours>\d+)h)?((?P<minutes>\d{1,2})m?)?$'
)


def must_be_datetime_or_relative(dt):
    """FIXME: Document"""
    if isinstance(dt, datetime.datetime):
        # FIXME: (lb): I've got milliseconds in my store data!!
        #        So this little hack kludge-fixes the problem;
        #        perhaps someday I'll revisit this and really
        #        figure out what's going on.
        return dt.replace(microsecond=0)
    elif not dt or isinstance(dt, str):
        if (
            dt and not (
                # NOTE: re.match checks for a match only at the beginning of the string.
                RE_PATTERN_RELATIVE_CLOCK.match(dt)
                or RE_PATTERN_RELATIVE_DELTA.match(dt)
            )
        ):
            raise TypeError(_(
                'Expected Hamster time string to indicate relative time, not: {}'
            ).format(dt))
        return dt
    raise TypeError(_(
        "Hamster time not `datetime`, relative string, or `None`, but: {}"
    ).format(type(dt)))


def must_not_start_after_end(range_tuple):
    """
    Perform basic sanity checks on a timeframe.

    Args:
        range_tuple (tuple): ``(start, end)`` tuple.

    Raises:
        ValueError: If start > end.

    Returns:
        tuple: ``(start, end)`` tuple that passed validation.

    Note:
        ``timeframes`` may be incomplete, e.g., end might not be set.
    """

    start, end = range_tuple

    if (
        isinstance(start, datetime.datetime)
        and isinstance(end, datetime.datetime)
        and start > end
    ):
        raise ValueError(_("Start after end!"))

    return range_tuple




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest

from nark.backends.sqlalchemy.objects import (
    AlchemyActivity,
    AlchemyCategory,
    AlchemyFact,
    AlchemyTag
)
from nark.backends.sqlalchemy.storage import SQLAlchemyStore
from nark.config import decorate_config


# The reason we see a great deal of count == 0 statements is to make sure that
# db rollback works as expected. Once we are confident in our sqlalchemy/pytest
# setup those are not really needed.
class TestStore(object):
    """Tests to make sure our store/test setup behaves as expected."""
    def test_build_is_not_persistent(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that calling ``factory.build()`` does not create a
        persistent db entry.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        alchemy_category_factory.build()
        assert alchemy_store.session.query(AlchemyCategory).count() == 0

    def test_factory_call_persistent(self, alchemy_store, alchemy_category_factory):
        """Make sure that ``factory()`` does creates a persistent db entry."""
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        alchemy_category_factory()
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_create_is_persistent(self, alchemy_store, alchemy_category_factory):
        """Make sure that  ``create()`` does creates a persistent db entry."""
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        alchemy_category_factory()
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_build_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure that factory instances have no pk assigned."""
        instance = alchemy_category_factory.build()
        assert instance.pk

    def test_create_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure that factory.create instances have pk assigned."""
        instance = alchemy_category_factory.create()
        assert instance.pk

    def test_instance_fixture(self, alchemy_store, alchemy_category):
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert alchemy_category.pk
        assert alchemy_category.name

    def test_get_db_url(self, alchemy_config_parametrized, alchemy_store):
        """Make sure that db_url composition works as expected."""
        config, expectation = alchemy_config_parametrized
        alchemy_store.config = decorate_config(config)
        assert alchemy_store.db_url == expectation

    def test_get_db_url_missing_keys(
        self, alchemy_config_missing_store_config_parametrized, alchemy_store,
    ):
        """
        Make sure that db_url composition throws error if key/values are
        missing in config.
        """
        with pytest.raises(ValueError):
            alchemy_store.config = decorate_config(
                alchemy_config_missing_store_config_parametrized
            )
            # If decorate_config() does not raise on engine missing error,
            # db_url will raise on missing path, host, etc.
            alchemy_store.db_url

    def test_init_with_unicode_path(self, alchemy_config, db_path_parametrized):
        """Test that Instantiating a store with a unicode path works."""
        alchemy_config['db.path'] = db_path_parametrized
        assert SQLAlchemyStore(alchemy_config)


class TestCategoryManager():
    def test_add_new(self, alchemy_store, alchemy_category_factory):
        """
        Our manager methods return the persistant instance as hamster objects.
        As we want to make sure that we compare our expectations against the
        raw saved object, we look it up explicitly again.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        category = alchemy_category_factory.build().as_hamster(alchemy_store)
        category.pk = None
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        result = alchemy_store.categories._add(category)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        db_instance = alchemy_store.session.query(AlchemyCategory).get(result.pk)
        assert category.equal_fields(db_instance)
        assert category != db_instance

    def test_add_existing_name(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that adding a category with a name that is already present
        gives an error.
        """
        existing_category = alchemy_category_factory()
        category = alchemy_category_factory.build().as_hamster(alchemy_store)
        category.name = existing_category.name
        category.pk = None
        with pytest.raises(ValueError):
            alchemy_store.categories._add(category)

    def test_add_with_pk(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that adding a alchemy_category that already got an PK
        raises an exception.
        """
        category = alchemy_category_factory().as_hamster(alchemy_store)
        category.name += 'foobar'
        assert category.pk
        with pytest.raises(ValueError):
            alchemy_store.categories._add(category)

    def test_update(
        self, alchemy_store, alchemy_category_factory, new_category_values,
    ):
        """Test that updateing a alchemy_category works as expected."""
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        # (lb): NOTE_TO_SELF: The alchemy_category_factory fixture was created
        # when conftest.py called `register(lib_factories.CategoryFactory)`,
        # which, AFAICT, is the same as calling the factory directly:
        #   alchemy_category_factory = factories.AlchemyCategoryFactory
        category = alchemy_category_factory().as_hamster(alchemy_store)
        new_values = new_category_values(category)
        for key, value in new_values.items():
            assert getattr(category, key) != value
        for key, value in new_values.items():
            setattr(category, key, value)
        alchemy_store.categories._update(category)
        db_instance = alchemy_store.session.query(AlchemyCategory).get(category.pk)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert category.equal_fields(db_instance)

    def test_update_without_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure that passing a category without a PK raises an error."""
        category = alchemy_category_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.categories._update(category)

    def test_update_invalid_pk(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that passing a category with a non existing PK raises an error.
        """
        category = alchemy_category_factory().as_hamster(alchemy_store)
        category.pk = category.pk + 10
        with pytest.raises(KeyError):
            alchemy_store.categories._update(category)

    def test_update_existing_name(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that renaming a given alchemy_category to a taken name
        throws an error.
        """
        category_1, category_2 = (
            alchemy_category_factory(), alchemy_category_factory(),
        )
        category_2 = category_2.as_hamster(alchemy_store)
        category_2.name = category_1.name
        with pytest.raises(ValueError):
            alchemy_store.categories._update(category_2)

    def test_remove(self, alchemy_store, alchemy_category_factory):
        """Make sure passing a valid alchemy_category removes it from the db."""
        alchemy_category = alchemy_category_factory()
        category = alchemy_category.as_hamster(alchemy_store)
        result = alchemy_store.categories.remove(category)
        assert result is None
        assert alchemy_store.session.query(AlchemyCategory).get(category.pk) is None

    def test_remove_no_pk(self, alchemy_store, alchemy_category_factory):
        """Ensure that passing a alchemy_category without an PK raises an error."""
        category = alchemy_category_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.categories.remove(category)

    def test_remove_invalid_pk(self, alchemy_store, alchemy_category_factory):
        """Ensure that passing a alchemy_category without an PK raises an error."""
        category = alchemy_category_factory.build(pk=800).as_hamster(alchemy_store)
        with pytest.raises(KeyError):
            alchemy_store.categories.remove(category)

    def test_get_existing_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure method retrieves corresponding object."""
        category = alchemy_category_factory().as_hamster(alchemy_store)
        result = alchemy_store.categories.get(category.pk)
        assert result == category

    def test_get_non_existing_pk(self, alchemy_store, alchemy_category_factory):
        """Make sure we throw an error if PK can not be resolved."""
        alchemy_store.session.query(AlchemyCategory).count == 0
        category = alchemy_category_factory()
        alchemy_store.session.query(AlchemyCategory).count == 1
        with pytest.raises(KeyError):
            alchemy_store.categories.get(category.pk + 1)

    def test_get_by_name(self, alchemy_category_factory, alchemy_store):
        """Make sure a alchemy_category can be retrieved by name."""
        category = alchemy_category_factory().as_hamster(alchemy_store)
        result = alchemy_store.categories.get_by_name(category.name)
        assert result == category

    def test_get_all(self, alchemy_store, set_of_categories):
        result = alchemy_store.categories.get_all()
        assert len(result) == len(set_of_categories)
        assert len(result) == alchemy_store.session.query(AlchemyCategory).count()
        for category in set_of_categories:
            assert category.as_hamster(alchemy_store) in result

    # Test convenience methods.
    def test_get_or_create_get(self, alchemy_store, alchemy_category_factory):
        """
        Test that if we pass a alchemy_category of existing name, we just
        return it.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        category = alchemy_category_factory().as_hamster(alchemy_store)
        result = alchemy_store.categories.get_or_create(category)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert result == category

    def test_get_or_create_new_name(self, alchemy_store, alchemy_category_factory):
        """
        Make sure that passing a category with new name creates and returns
        new instance.
        """
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        category = alchemy_category_factory.build().as_hamster(alchemy_store)
        category.pk = None
        result = alchemy_store.categories.get_or_create(category)
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert result.equal_fields(category)


class TestActivityManager():
    def test_get_or_create_get(self, alchemy_store, alchemy_activity):
        """
        Make sure that passing an existing activity retrieves the corresponding
        instance.

        Note:
            * The activity will is be looked up by its composite key, so not to
            make any assumptions on the existence of a PK.
        """
        activity = alchemy_activity.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        result = alchemy_store.activities.get_or_create(activity)
        assert result == activity
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_get_or_create_new(self, alchemy_store, activity):
        """
        Make sure that passing a new activity create a new persitent instance.

        Note:
            * The activity will is be looked up by its composite key, so not to
            make any assumptions on the existence of a PK.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        result = alchemy_store.activities.get_or_create(activity)
        assert result.equal_fields(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_save_new(self, activity, alchemy_store):
        """Make sure that saving a new activity add a new persistent instance."""
        # [TODO]
        # This should not be needed as ``save`` is a basestore method.
        # Its just a case of 'better save than sorry.
        assert activity.pk is None
        count_before = alchemy_store.session.query(AlchemyActivity).count()
        result = alchemy_store.activities._add(activity)
        count_after = alchemy_store.session.query(AlchemyActivity).count()
        assert count_before < count_after
        assert result.equal_fields(activity)

    def test_save_existing(
        self, alchemy_store, alchemy_activity, alchemy_category_factory,
    ):
        """
        Make sure that saving an existing activity add no new persistent
        instance.
        """
        # [TODO]
        # This should not be needed as ``save`` is a basestore method.
        activity = alchemy_activity.as_hamster(alchemy_store)
        activity.category = alchemy_category_factory()
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2
        result = alchemy_store.activities.save(activity)
        assert result == activity
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2

    def test_activity_without_category(self, alchemy_store, activity):
        """Add a new activity without an category."""
        activity.category = None
        result = alchemy_store.activities._add(activity)
        assert result.equal_fields(activity)

    def test_add_new_with_new_category(self, alchemy_store, activity, category):
        """
        Test that adding a new alchemy_activity with new alchemy_category
        creates both.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert alchemy_store.session.query(AlchemyCategory).count() == 0
        activity.category = category
        result = alchemy_store.activities._add(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_add_new_with_existing_category(
        self, alchemy_store, activity, alchemy_category,
    ):
        """
        Test that adding a new activity with existing category does not
        create a new one.
        """
        activity.category = alchemy_category.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        result = alchemy_store.activities._add(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_add_new_with_existing_name_and_alchemy_category(
        self, alchemy_store, activity, alchemy_activity,
    ):
        """
        Test that adding a new alchemy_activity_with_existing_composite_key
        throws error.
        """
        activity.name = alchemy_activity.name
        activity.category = alchemy_activity.category.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        with pytest.raises(ValueError):
            alchemy_store.activities._add(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1

    def test_add_with_pk(self, alchemy_store, activity):
        """Make sure that adding an alchemy_activity with a PK raises error."""
        activity.pk = 234
        with pytest.raises(ValueError):
            alchemy_store.activities._add(activity)

    def test_update_without_pk(self, alchemy_store, activity):
        """Make sure that calling update without a PK raises exception."""
        with pytest.raises(ValueError):
            alchemy_store.activities._update(activity)

    def test_update_with_existing_name_and_existing_category_name(
        self, alchemy_store, activity, alchemy_activity, alchemy_category_factory,
    ):
        """
        Make sure that calling update with a taken composite key raises
        exception.
        """
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 1
        category = alchemy_category_factory()
        assert alchemy_activity.category != category
        activity.name = alchemy_activity.name
        assert activity.category.pk is None
        activity.category.name = category.name
        with pytest.raises(ValueError):
            alchemy_store.activities._update(activity)

    def test_update_with_existing_category(
        self, alchemy_store, alchemy_activity, alchemy_category_factory,
    ):
        """
        Test that updateting an activity with existing category does not
        create a new one.
        """
        activity = alchemy_activity.as_hamster(alchemy_store)
        category = alchemy_category_factory().as_hamster(alchemy_store)
        assert alchemy_activity.category != category
        activity.category = category
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2
        result = alchemy_store.activities._update(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert alchemy_store.session.query(AlchemyCategory).count() == 2
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_update_name(
        self, alchemy_store, alchemy_activity, name_string_valid_parametrized,
    ):
        """Test updateing an activities name with a valid new string."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        activity.name = name_string_valid_parametrized
        result = alchemy_store.activities._update(activity)
        db_instance = alchemy_store.session.query(AlchemyActivity).get(result.pk)
        assert db_instance.as_hamster(alchemy_store).equal_fields(activity)

    def test_remove_existing(self, alchemy_store, alchemy_activity):
        """Make sure removing an existsing alchemy_activity works as intended."""
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        activity = alchemy_activity.as_hamster(alchemy_store)
        result = alchemy_store.activities.remove(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        assert result is True

    def test_remove_no_pk(self, alchemy_store, activity):
        """
        Make sure that trying to remove an alchemy_activity without a PK
        raises error.
        """
        with pytest.raises(ValueError):
            alchemy_store.activities.remove(activity)

    def test_remove_invalid_pk(self, alchemy_store, alchemy_activity):
        """Test that removing of a non-existent key raises error."""
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        activity = alchemy_activity.as_hamster(alchemy_store)
        activity.pk = activity.pk + 1
        with pytest.raises(KeyError):
            alchemy_store.activities.remove(activity)
        assert alchemy_store.session.query(AlchemyActivity).count() == 1

    def test_get_existing(self, alchemy_store, alchemy_activity):
        """
        Make sure that retrieving an existing alchemy_activity by pk works
        as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity.pk)
        assert result == alchemy_activity.as_hamster(alchemy_store)
        assert result is not alchemy_activity

    def test_get_existing_raw(self, alchemy_store, alchemy_activity):
        """
        Make sure that retrieving an existing alchemy_activity by pk works
        as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity.pk, raw=True)
        assert result == alchemy_activity
        assert result is alchemy_activity

    # FIXME: Add deleted/hidden tests...
    # FIXME: Split this file into one file for each item type.
    def test_get_existing_deleted_false(self, alchemy_store, alchemy_activity):
        """
        Make sure that retrieving an existing alchemy_activity by
        deleted=False works as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity.pk, deleted=False)
        assert result == alchemy_activity.as_hamster(alchemy_store)
        assert result is not alchemy_activity

    def test_get_existing_deleted_true(self, alchemy_store, alchemy_activity_deleted):
        """
        Make sure that retrieving an existing alchemy_activity by
        deleted=True works as intended.
        """
        result = alchemy_store.activities.get(alchemy_activity_deleted.pk, deleted=True)
        assert result == alchemy_activity_deleted.as_hamster(alchemy_store)
        assert result is not alchemy_activity_deleted

    def test_get_non_existing(self, alchemy_store):
        """Make sure quering for a non existent PK raises error."""
        with pytest.raises(KeyError):
            alchemy_store.activities.get(4)

    @pytest.mark.parametrize('raw', (True, False))
    def test_get_by_composite_valid(self, alchemy_store, alchemy_activity, raw):
        """Make sure that querying for a valid name/alchemy_category combo succeeds."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        result = alchemy_store.activities.get_by_composite(
            activity.name,
            activity.category,
            raw=raw,
        )
        if raw:
            assert result == alchemy_activity
            assert result is alchemy_activity
        else:
            assert result == alchemy_activity
            assert result is not alchemy_activity

    def test_get_by_composite_invalid_category(
        self, alchemy_store, alchemy_activity, alchemy_category_factory,
    ):
        """Make sure that querying with an invalid category raises errror."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        category = alchemy_category_factory().as_hamster(alchemy_store)
        with pytest.raises(KeyError):
            alchemy_store.activities.get_by_composite(activity.name, category)

    def test_get_by_composite_invalid_name(
        self, alchemy_store, alchemy_activity, name_string_valid_parametrized,
    ):
        """Make sure that querying with an invalid alchemy_category raises errror."""
        activity = alchemy_activity.as_hamster(alchemy_store)
        invalid_name = activity.name + 'foobar'
        with pytest.raises(KeyError):
            alchemy_store.activities.get_by_composite(invalid_name, activity.category)

    def test_get_all_without_category(self, alchemy_store, alchemy_activity):
        """Make sure method returns all activities."""
        result = alchemy_store.activities.get_all()
        assert len(result) == 1

    def test_get_all_with_category_none(
        self, alchemy_store, alchemy_activity, alchemy_activity_factory,
    ):
        """Make sure only activities without a category are are returned."""
        # FIXME: Do we care about _activity? result ==/is _activity ?
        #   activity = alchemy_activity_factory(category=None)
        alchemy_activity_factory(category=None)
        result = alchemy_store.activities.get_all(category=alchemy_activity.category)
        assert len(result) == 1

    def test_get_all_with_category(self, alchemy_store, alchemy_activity):
        """
        Make sure that activities matching the given alchemy_category are returned.
        """
        activity = alchemy_activity.as_hamster(alchemy_store)
        result = alchemy_store.activities.get_all(category=activity.category)
        assert len(result) == 1

    def test_get_all_with_search_term(self, alchemy_store, alchemy_activity):
        """
        Make sure that activities matching the given term ass name are returned.
        """
        activity = alchemy_activity.as_hamster(alchemy_store)
        result = alchemy_store.activities.get_all(
            category=activity.category,
            search_term=activity.name,
        )
        assert len(result) == 1


class TestTagManager():
    def test_add_new(self, alchemy_store, alchemy_tag_factory):
        """
        Our manager methods return the persistant instance as hamster objects.
        As we want to make sure that we compare our expectations against the
        raw saved object, we look it up explicitly again.
        """
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory.build().as_hamster(alchemy_store)
        tag.pk = None
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        result = alchemy_store.tags._add(tag)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        db_instance = alchemy_store.session.query(AlchemyTag).get(result.pk)
        assert tag.equal_fields(db_instance)
        assert tag != db_instance

    def test_add_existing_name(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that adding a tag with a name that is already present
        gives an error.
        """
        existing_tag = alchemy_tag_factory()
        tag = alchemy_tag_factory.build().as_hamster(alchemy_store)
        tag.name = existing_tag.name
        tag.pk = None
        with pytest.raises(ValueError):
            alchemy_store.tags._add(tag)

    def test_add_with_pk(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that adding a alchemy_tag that already got an PK raises
        an exception.
        """
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        tag.name += 'foobar'
        assert tag.pk
        with pytest.raises(ValueError):
            alchemy_store.tags._add(tag)

    def test_update(self, alchemy_store, alchemy_tag_factory, new_tag_values):
        """Test that updateing a alchemy_tag works as expected."""
        alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        new_values = new_tag_values(tag)
        for key, value in new_values.items():
            assert getattr(tag, key) != value
        for key, value in new_values.items():
            setattr(tag, key, value)
        alchemy_store.tags._update(tag)
        db_instance = alchemy_store.session.query(AlchemyTag).get(tag.pk)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        assert tag.equal_fields(db_instance)

    def test_update_without_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure that passing a tag without a PK raises an error."""
        tag = alchemy_tag_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.tags._update(tag)

    def test_update_invalid_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure that passing a tag with a non existing PK raises an error."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        tag.pk = tag.pk + 10
        with pytest.raises(KeyError):
            alchemy_store.tags._update(tag)

    def test_update_existing_name(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that renaming a given alchemy_tag to a taken name throws an error.
        """
        tag_1, tag_2 = (alchemy_tag_factory(), alchemy_tag_factory())
        tag_2 = tag_2.as_hamster(alchemy_store)
        tag_2.name = tag_1.name
        with pytest.raises(ValueError):
            alchemy_store.tags._update(tag_2)

    def test_remove(self, alchemy_store, alchemy_tag_factory):
        """Make sure passing a valid alchemy_tag removes it from the db."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.remove(tag)
        assert result is None
        assert alchemy_store.session.query(AlchemyTag).get(tag.pk) is None

    def test_remove_no_pk(self, alchemy_store, alchemy_tag_factory):
        """Ensure that passing a alchemy_tag without an PK raises an error."""
        tag = alchemy_tag_factory.build(pk=None).as_hamster(alchemy_store)
        with pytest.raises(ValueError):
            alchemy_store.tags.remove(tag)

    def test_remove_invalid_pk(self, alchemy_store, alchemy_tag_factory):
        """Ensure that passing a alchemy_tag without an PK raises an error."""
        tag = alchemy_tag_factory.build(pk=800).as_hamster(alchemy_store)
        with pytest.raises(KeyError):
            alchemy_store.tags.remove(tag)

    def test_get_existing_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure method retrieves corresponding object."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.get(tag.pk)
        assert result == tag

    def test_get_non_existing_pk(self, alchemy_store, alchemy_tag_factory):
        """Make sure we throw an error if PK can not be resolved."""
        alchemy_store.session.query(AlchemyTag).count == 0
        tag = alchemy_tag_factory()
        alchemy_store.session.query(AlchemyTag).count == 1
        with pytest.raises(KeyError):
            alchemy_store.tags.get(tag.pk + 1)

    def test_get_by_name(self, alchemy_tag_factory, alchemy_store):
        """Make sure a alchemy_tag can be retrieved by name."""
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.get_by_name(tag.name)
        assert result == tag

    def test_get_all(self, alchemy_store, set_of_tags):
        result = alchemy_store.tags.get_all()
        assert len(result) == len(set_of_tags)
        assert len(result) == alchemy_store.session.query(AlchemyTag).count()
        for tag in set_of_tags:
            assert tag.as_hamster(alchemy_store) in result

    # Test convenience methods.
    def test_get_or_create_get(self, alchemy_store, alchemy_tag_factory):
        """
        Test that if we pass a alchemy_tag of existing name, we just return it.
        """
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory().as_hamster(alchemy_store)
        result = alchemy_store.tags.get_or_create(tag)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        assert result == tag

    def test_get_or_create_new_name(self, alchemy_store, alchemy_tag_factory):
        """
        Make sure that passing a tag with new name creates and returns new instance.
        """
        assert alchemy_store.session.query(AlchemyTag).count() == 0
        tag = alchemy_tag_factory.build().as_hamster(alchemy_store)
        tag.pk = None
        result = alchemy_store.tags.get_or_create(tag)
        assert alchemy_store.session.query(AlchemyTag).count() == 1
        assert result.equal_fields(tag)


class TestFactManager():
    def test_timeframe_available_existing_fact_overlaps_start_only(
        self, alchemy_store, fact, alchemy_fact,
    ):
        """
        Make sure that passing a fact with only start overlapping an existing
        one raises error.
        """
        fact.start = alchemy_fact.start - datetime.timedelta(days=4)
        fact.end = alchemy_fact.start + datetime.timedelta(minutes=15)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    def test_timeframe_available_existing_fact_overlaps_end_only(
        self, alchemy_store, fact, alchemy_fact,
    ):
        """
        Make sure that passing a fact with only end overlapping an existing
        one raises error.
        """
        fact.start = alchemy_fact.end - datetime.timedelta(minutes=1)
        fact.end = alchemy_fact.end + datetime.timedelta(minutes=15)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    # Testcase for Bug LIB-253
    def test_timeframe_available_fact_completely_within_existing_timeframe(
        self, alchemy_store, fact, alchemy_fact,
    ):
        """
        Make sure that passing a fact that is completely within an existing
        ones raises error.
        """
        fact.start = alchemy_fact.start + datetime.timedelta(minutes=1)
        fact.end = alchemy_fact.end - datetime.timedelta(minutes=1)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    def test_timeframe_available_existing_fact_completly_spans_existing_timeframe(
        self, alchemy_store, fact, alchemy_fact,
    ):
        """
        Make sure that passing a fact that completely spans an existing fact
        raises an error.
        """
        fact.start = alchemy_fact.start - datetime.timedelta(minutes=1)
        fact.end = alchemy_fact.end + datetime.timedelta(minutes=15)
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    def test_add_tags(self, alchemy_store, fact):
        """Make sure that adding a new valid fact will also save its tags."""
        result = alchemy_store.facts._add(fact)
        assert fact.tags
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert db_instance.tags
        assert db_instance.as_hamster(alchemy_store).equal_fields(fact)

    def test_add_new_valid_fact_new_activity(self, alchemy_store, fact):
        """
        Make sure that adding a new valid fact with a new activity works as
        intended.
        """
        assert alchemy_store.session.query(AlchemyFact).count() == 0
        assert alchemy_store.session.query(AlchemyActivity).count() == 0
        result = alchemy_store.facts._add(fact)
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert alchemy_store.session.query(AlchemyFact).count() == 1
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(fact)

    def test_add_new_valid_fact_existing_activity(
        self, alchemy_store, fact, alchemy_activity,
    ):
        """
        Make sure that adding a new valid fact with an existing activity works
        as intended.
        """
        fact.activity = alchemy_activity.as_hamster(alchemy_store)
        assert alchemy_store.session.query(AlchemyFact).count() == 0
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        result = alchemy_store.facts._add(fact)
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert alchemy_store.session.query(AlchemyFact).count() == 1
        assert alchemy_store.session.query(AlchemyActivity).count() == 1
        assert db_instance.as_hamster(alchemy_store).equal_fields(fact)

    def test_add_with_pk(self, alchemy_store, fact):
        """Make sure that passing a fact with a PK raises error."""
        fact.pk = 101
        with pytest.raises(ValueError):
            alchemy_store.facts._add(fact)

    def test_update_respects_tags(self, alchemy_store, alchemy_fact, new_fact_values):
        """Make sure that updating sets tags as expected."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        new_values = new_fact_values(fact)
        fact.tags = new_values['tags']
        result = alchemy_store.facts._update(fact)
        # Because split_from, fact will have been marked deleted.
        assert fact.deleted
        # And the new Fact will have a larger PK.
        # (lb): We really don't need to impose a strictly increasing order
        # property on the PK, but it's how our code happens to work, so might
        # as well check it.
        assert result.pk > fact.pk
        fact.deleted = False
        assert result.split_from.pk == fact.pk
        # Note that the split-from is not the same because it contains previous tags.
        assert result.split_from != fact
        db_instance = alchemy_store.session.query(AlchemyFact).get(result.pk)
        assert db_instance.as_hamster(alchemy_store).equal_fields(result)
        result.split_from = None
        assert result.equal_fields(fact)

    def test_update_nonexisting_fact(self, alchemy_store, alchemy_fact, new_fact_values):
        """Make sure that trying to update a fact that does not exist raises error."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        new_values = new_fact_values(fact)
        fact.start = new_values['start']
        fact.end = new_values['end']
        fact.pk += 100
        with pytest.raises(KeyError):
            alchemy_store.facts._update(fact)

    def test_update_new_fact(self, alchemy_store, alchemy_fact, new_fact_values):
        """Make sure that trying to update a new fact ,e.g. one without a pk."""
        fact = alchemy_fact.as_hamster(alchemy_store)
        new_values = new_fact_values(fact)
        fact.start = new_values['start']
        fact.end = new_values['end']
        fact.pk = None
        with pytest.raises(ValueError):
            alchemy_store.facts._update(fact)

    def test_save_new(self, fact, alchemy_store):
        count_before = alchemy_store.session.query(AlchemyFact).count()
        result = alchemy_store.facts.save(fact)
        count_after = alchemy_store.session.query(AlchemyFact).count()
        assert count_before < count_after
        assert result.activity.name == fact.activity.name
        assert result.description == fact.description

    def test_remove(self, alchemy_store, alchemy_fact):
        """Make sure the fact but not its tags are removed."""
        count_before = alchemy_store.session.query(AlchemyFact).count()
        tags_before = alchemy_store.session.query(AlchemyTag).count()
        fact = alchemy_fact.as_hamster(alchemy_store)
        result = alchemy_store.facts.remove(fact, purge=True)
        count_after = alchemy_store.session.query(AlchemyFact).count()
        assert count_after < count_before
        assert result is True
        assert alchemy_store.session.query(AlchemyFact).get(fact.pk) is None
        assert alchemy_store.session.query(AlchemyTag).count() == tags_before

    def test_get(self, alchemy_store, alchemy_fact):
        fact = alchemy_fact.as_hamster(alchemy_store)
        result = alchemy_store.facts.get(fact.pk)
        assert result == fact

    def test_get_all(self, set_of_alchemy_facts, alchemy_store):
        result = alchemy_store.facts._get_all()
        assert len(result) == len(set_of_alchemy_facts)
        assert len(result) == alchemy_store.session.query(AlchemyFact).count()

    @pytest.mark.parametrize(('start_filter', 'end_filter'), (
        (10, 12),
        (10, None),
        (None, -12),
    ))
    def test_get_all_existing_facts_not_in_timerange(
        self,
        alchemy_store,
        alchemy_fact,
        bool_value_parametrized,
        start_filter,
        end_filter,
    ):
        """Make sure that a valid timeframe returns an empty list."""
        since, until = None, None
        if start_filter:
            since = alchemy_fact.start + datetime.timedelta(days=start_filter)
        if end_filter:
            until = alchemy_fact.start + datetime.timedelta(days=end_filter)

        result = alchemy_store.facts._get_all(
            since=since, until=until, partial=bool_value_parametrized,
        )
        assert result == []

    @pytest.mark.parametrize(('start_filter', 'end_filter'), (
        (-1, 5),
        (-1, None),
        (None, 5),
        (None, None),
    ))
    def test_get_all_existing_fact_fully_in_timerange(
        self,
        alchemy_store,
        alchemy_fact,
        bool_value_parametrized,
        start_filter,
        end_filter,
    ):
        """Ensure a fact fully within the timeframe is returned."""
        since, until = None, None
        if start_filter:
            since = alchemy_fact.start + datetime.timedelta(days=start_filter)
        if end_filter:
            until = alchemy_fact.start + datetime.timedelta(days=end_filter)

        result = alchemy_store.facts._get_all(
            since=since, until=until, partial=bool_value_parametrized,
        )
        # ANSWER/2018-05-05: (lb): This test is failing. When did it break?
        #   assert result == [alchemy_fact]
        assert len(result) == 1
        assert str(result[0]) == str(alchemy_fact)

    @pytest.mark.parametrize(('start_filter', 'end_filter'), (
        # Fact.start is in timewindow
        (None, 2),
        (-900, 2),
        # Fact.end is in timewindow
        (5, None),
        (5, 900),
    ))
    def test_get_all_existing_fact_partialy_in_timerange(
        self,
        alchemy_store,
        alchemy_fact,
        bool_value_parametrized,
        start_filter,
        end_filter,
    ):
        """
        Test that a fact partially within timeframe is returned with
        ``partial=True`` only.
        """
        since, until = None, None
        if start_filter:
            since = alchemy_fact.start + datetime.timedelta(minutes=start_filter)
        if end_filter:
            until = alchemy_fact.start + datetime.timedelta(minutes=end_filter)
        result = alchemy_store.facts._get_all(
            since=since,
            until=until,
            partial=bool_value_parametrized,
            # 2019-02-14: (lb): When lazy_tags=False, the `result == [alchemy_fact]`
            # fails because result[0].tags != alchemy_fact.tags, because it seems
            # query.all() loads Tag objects with proper PKs, but lazy_tags uses
            # query-coalesce magic to assemble Tag labels without PKs, so PK=None.
            # FIXME/TESTME/2019-02-14: (lb): What's behavior on, say, dob-list?
            #   py.test --pdb -vv \
            #       -k test_get_all_existing_fact_partialy_in_timerange tests/
            lazy_tags=True,
        )
        if bool_value_parametrized:
            assert len(result) == 1
            assert str(result[0]) == str(alchemy_fact)
            assert result == [alchemy_fact]
        else:
            assert result == []

    def test_get_all_search_matches_activity(self, alchemy_store, set_of_alchemy_facts):
        """Make sure facts with ``Fact.activity.name`` matching the term are returned."""
        search_term = set_of_alchemy_facts[1].activity.name
        result = alchemy_store.facts._get_all(search_term=search_term, lazy_tags=True)
        assert len(result) == 1
        assert len(set_of_alchemy_facts) == 5
        assert str(result[0]) == str(set_of_alchemy_facts[1])
        assert result == [set_of_alchemy_facts[1]]

    def test_get_all_search_matches_category(self, alchemy_store, set_of_alchemy_facts):
        """Make sure facts with ``Fact.category.name`` matching the term are returned."""
        search_term = set_of_alchemy_facts[1].category.name
        result = alchemy_store.facts._get_all(search_term=search_term, lazy_tags=True)
        assert len(result) == 1
        assert len(set_of_alchemy_facts) == 5
        assert str(result[0]) == str(set_of_alchemy_facts[1])
        assert result == [set_of_alchemy_facts[1]]




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from . import BaseManager


class BaseMigrationsManager(BaseManager):
    """Base class defining the minimal API for a MigrationsManager implementation."""

    def downgrade(self):
        """Downgrade the database according to its migration version."""
        raise NotImplementedError

    def upgrade(self):
        """Upgrade the database according to its migration version."""
        raise NotImplementedError

    def version(self):
        """Returns the migration version of the database indicated by the config."""
        raise NotImplementedError

    def latest_version(self):
        """Returns the latest version of the database as used by the application."""
        raise NotImplementedError

    # ***

    def migrations_config(self):
        """Return sqlalchemy-migrate migrate.cfg as dictionary object."""
        raise NotImplementedError

    def migrations_path(self):
        """Return the path to the migration versions/ base directory."""
        raise NotImplementedError

    def legacy_upgrade_from_hamster_applet(self, db_path):
        """Upgrade legacy SQLite database created by hamster-applet."""
        raise NotImplementedError

    def legacy_upgrade_from_hamster_lib(self):
        """Upgrade legacy SQLite database created by hamster-lib."""
        raise NotImplementedError




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""
Base classes for implementing storage backends.

Note:
    * This is propably going to be replaced by a ``ABC``-bases solution.
    * Basic sanity checks could be done here then. This would mean we just need
      to test them once and our actual backends focus on the CRUD implementation.
"""

from gettext import gettext as _

from ..items.item_base import BaseItem

__all__ = ('BaseManager', )


class BaseManager(object):
    """Base class for all object managers."""

    def __init__(self, store):
        self.store = store

    @property
    def config(self):
        return self.store.config

    # ***

    def adding_item_must_not_have_pk(self, hamster_item):
        message = _("Adding item: {!r}.".format(hamster_item))
        self.store.logger.debug(message)
        if not hamster_item.pk:
            return
        message = _(
            "The {} item ({!r}) cannot be added because it already has a PK."
            " Perhaps call the ``_update`` method instead".format(
                self.__class__.__name__, hamster_item,
            )
        )
        self.store.logger.error(message)
        raise ValueError(message)

    # ***

    def save(self, item, cls=BaseItem, named=False, **kwargs):
        """
        Save a Nark object instance to user's selected backend.

        Will either ``_add`` or ``_update`` based on item PK.

        Args:
            tag (nark.BaseItem, i.e., Activity/Category/Fact/Tag):
                Nark instance to be saved.

        Returns:
            nark.BaseItem: Saved Nark instance.

        Raises:
            TypeError: If the ``item`` parameter is not a valid ``BaseItem`` instance.
        """

        if not isinstance(item, cls):
            message = _("You need to pass a {} object").format(cls.__name__)
            self.store.logger.debug(message)
            raise TypeError(message)

        # (lb): Not sure this is quite what we want, but Activity has been doing this,
        # and I just made this base class, so now all items will be doing this.
        if named and not item.name:
            raise ValueError(_("You must specify an item name."))

        self.store.logger.debug(_("'{}' has been received.".format(item)))

        # NOTE: Not assuming that PK is an int, i.e., not testing '> 0'.
        if item.pk or item.pk == 0:
            result = self._update(item, **kwargs)
        else:
            # PK is empty string, empty list, None, etc., but not 0.
            result = self._add(item, **kwargs)
        return result

